#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 20:41
import time

from pyxllib.prog.pupil import check_install_package

check_install_package('paramiko')
check_install_package('scp')
# 对 paramiko 进一步封装的库
# check_install_package('fabric')

from collections import defaultdict
import os
import pathlib
import re
import shutil
import sys

import humanfriendly
import pandas as pd
import paramiko
import scp as scplib
from tqdm import tqdm

from pyxllib.algo.pupil import natural_sort
from pyxllib.file.specialist import XlPath
from pyxllib.debug.specialist import get_xllog

logger = get_xllog('location')


class SshCommandError(Exception):
    pass


class ScpLimitError(Exception):
    pass


class ScpProgress:
    # 这里的scp类比较特别，要用特殊的方式来实现一个tqdm式的进度条
    def __init__(self, info, **kwargs):
        self.info = info
        self.kwargs = kwargs

        self.trans_size = 0  # 总共传输了多少
        self.this_file_sent = 0  # 当前文件传输了多少

        if info == 0:
            self.tqdm = None
        elif info == 1:
            self.kwargs['frequency'] = 0  # 传输频数
            self.tqdm = tqdm(desc=self.kwargs.get('desc'), disable=False,
                             total=self.kwargs['total'], unit_scale=True)
        else:
            raise NotImplementedError

    def __call__(self, filename, size, sent):
        # 1 记录更新信息
        if isinstance(filename, bytes):
            filename = filename.decode('utf8')
        if 'rf' in self.kwargs:  # 仅显示相对路径
            filename = self.kwargs['rf']
        elif 'desc' in self.kwargs and filename.startswith(self.kwargs["desc"][1:]):
            filename = filename[len(self.kwargs["desc"]):]

        finish_file = (size == sent)  # 传完一个文件
        increment = sent - self.this_file_sent  # 默认每次是上传16kb，可以在SCP初始化修改buff_size
        self.trans_size += increment
        self.this_file_sent = 0 if finish_file else sent

        # 2 不同info的处理
        if self.info == 0:
            pass
        elif self.info == 1:
            self.kwargs['frequency'] += 1
            self.tqdm.desc = f'{self.kwargs["desc"]} > {sent / size:4.0%} ' + filename
            self.tqdm.n += increment
            if finish_file or self.kwargs['frequency'] % 100 == 0:  # 减小展示间隔
                self.tqdm.update(0)

        # 3 检查limit_bytes
        if sent == size and self.kwargs.get('limit_bytes'):
            if self.trans_size >= self.kwargs.get('limit_bytes'):
                if self.info:
                    logger.warning('达到限定上传大小，早停。')
                raise ScpLimitError


class XlSSHClient(paramiko.SSHClient):
    def __1_初始化和执行(self):
        pass

    def __init__(self, server, port, user, passwd, *, map_path=None,
                 relogin=0, relogin_interval=1):
        """

        :param map_path: 主要在上传、下载文件的时候，可以用来自动定位路径
            参考写法：{'C:/': '/'}  # 将D盘映射到服务器位置
        :param int relogin: 当连接失败的时候，可能只是某种特殊原因，可以尝试重新连接
            该参数设置重连的次数
        :param int|float relogin_interval: 每次重连的间隔秒数

        """

        super().__init__()
        self.load_system_host_keys()
        self.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # 重试登录
        for k in range(relogin + 1):
            try:
                self.connect(server, port, user, passwd)
                break
            except paramiko.ssh_exception.SSHException:
                if k < relogin:
                    time.sleep(relogin_interval)
        else:
            raise paramiko.ssh_exception.SSHException

        self.map_path = map_path

        class Path(pathlib.PurePosixPath):
            """ 生成一个可以管理远程服务器上的路径类

            该类只是在确实有单文件判断需求下方便使用
            实际处理问题中，推荐优先使用多文件一次性统一处理的命令，减小和服务器间的交互，提高运行速度
            """
            client = self

            def exists_type(self):
                # TODO 还没有考虑符号链接等情况，在特殊情况下可能会出意外
                t = self.client.exec(f'if test -d "{self}"; then echo -1;'
                                     f'elif test -f "{self}"; then echo 1;'
                                     'else echo 0; fi')
                return int(t)

            def sub_rel_paths(self, mtime=False, *, extcmd=''):
                """ 目录对象的时候，可以获取远程目录下的文件清单（相对路径）

                :param extcmd: 一般是增加 -maxdepth 1 的参数，限制只遍历第1层
                :param mtime: 增加返回mtime信息
                """
                printf = '%P %Ts\n' if mtime else r'%P\n'
                cmd = f'find "{self}" -mindepth 1'
                if extcmd: cmd += ' ' + extcmd
                cmd += f' -printf "{printf}"'
                stdout = self.client.exec(cmd)
                lines = stdout.splitlines()
                if mtime:
                    lines = [list(re.match(r'(.+?)\s(\d+)$', line).groups()) for line in lines]
                    return {line[0]: int(line[1]) for line in lines}
                else:
                    return set(lines)

            def mtime(self):
                res = self.client.exec(f'find "{self}" -maxdepth 0 -printf "%Ts"')
                return int(res)

            def set_mtime(self, timestamp):
                """ 手动修改文件的mtime

                我也不想搞这么复杂，但是scp底层传输时间戳有误差，没办法啊~~~
                """
                pass

            def size(self, *, human_readable=False):
                # 其实这里用-h也能美化，但怕计量方式跟XlPath.size不同，所以还是用humanfriendly再算
                try:
                    res = self.client.exec(f'du "{self}" -s -b')
                except SshCommandError:
                    res = 0

                sz = int(re.match(r'\d+', res).group())
                if human_readable:
                    return humanfriendly.format_size(self.size, binary=True)
                else:
                    return sz

        self.Path = Path

    @classmethod
    def login(cls, host_name, user_name, map_path=None, **kwargs):
        r""" 使用XlprDb里存储的服务器、账号信息，进行登录
        """
        from pyxllib.data.pglib import XlprDb

        with XlprDb.connect() as con:
            if host_name.startswith('g_'):
                host_ip = con.execute("SELECT host_ip FROM hosts WHERE host_name='xlpr0'").fetchone()[0]
                pw, port = con.execute('SELECT (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text, frpc_port'
                                       ' FROM hosts WHERE host_name=%s',
                                       (con.seckey, user_name, host_name[2:])).fetchone()
            else:
                port = 22
                host_ip, pw = con.execute('SELECT host_ip, (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text'
                                          ' FROM hosts WHERE host_name=%s',
                                          (con.seckey, user_name, host_name)).fetchone()

        if map_path is None:
            if sys.platform == 'win32':
                map_path = {'C:/': '/'}
            else:
                map_path = {'/': '/'}
        return cls(host_ip, port, user_name, pw[1:-1], map_path=map_path, **kwargs)

    def exec(self, command, *args, **kwargs):
        """ exec_command的简化版

        如果stderr出错，则抛出异常
        否则返回运行结果的文本数据

        【备忘】
        nginx -t的两句返回，虽然是正确状态，默认是放在stderr的
        """
        stdin, stdout, stderr = self.exec_command(command, *args, **kwargs)
        stderr = list(stderr)
        if stderr:  # TODO 目前警告也会报错，其实警告没关系
            print(''.join(stderr))
            raise SshCommandError(f'服务器执行命令报错: {command}')
        return '\n'.join([f.strip() for f in list(stdout)])

    def exec_script(self, main_cmd, script='', *, file=None):
        r""" 执行批量、脚本命令

        :paramn main_cmd: 主命令
        :param script: 用字符串表达的一套命令
        :param file: 也可以直接传入一个文件

        【使用示例】
        ssh = XlSSHClient.log_in('xlpr10', 'chenkunze')
        text = textwrap.dedent('''\
        import os
        print(len(os.listdir()))
        ''')
        print(ssh.exec_script('python3', text))
        """
        # 1 将脚本转成文件，上传到服务器
        scp = scplib.SCPClient(self.get_transport())

        # 虽然是基于本地的情况生成随机名称脚本，但一般在服务器也不会冲突，概率特别小
        local_file = XlPath.tempfile()
        host_file = '/tmp/' + local_file.name

        if file is not None:
            shutil.copy2(XlPath(file), local_file)
        elif script:
            local_file.write_text(script)
        else:
            raise ValueError(f'没有待执行的脚本')

        scp.put(local_file, host_file)
        local_file.delete()

        # 2 执行程序
        res = self.exec(f'{main_cmd} {host_file}')
        self.exec(f'rm {host_file}')
        return res

    def __2_scp(self):
        """ 以下是为scp准备的功能 """
        pass

    def __local_dir(self, remote_path, local_dir):
        if local_dir is not None:
            return XlPath(local_dir)
        else:
            for k, v in self.map_path.items():
                try:
                    relpath = remote_path.relative_to(v)
                    return (XlPath(k) / relpath).parent
                except ValueError:
                    pass
            raise ValueError('找不到对应的map_path路径映射规则')

    def __remote_dir(self, local_path, remote_dir):
        if remote_dir is not None:
            return self.Path(remote_dir)
        else:
            for k, v in self.map_path.items():
                try:
                    relpath = local_path.relative_to(k)
                    return (self.Path(v) / relpath.as_posix()).parent
                except ValueError:
                    pass
            raise ValueError('找不到对应的map_path路径映射规则')

    def __check_filetype(self, local_path, remote_path):
        """ 同类型或有一边不存在都可以

        不存在记为0，文件记为1，目录记为-1，那么只要乘积不为负数即可。

        尽量少用这个，毕竟这个要连接服务器，是有资源开销的。
        """
        a = local_path.exists_type()
        b = remote_path.exists_type()
        if a * b < 0:
            raise TypeError(f'本地和服务器的文件类型不匹配 {local_path} {remote_path}')
        return a, b

    def __scp_base(self, func, progress, from_path, to_dir, to_path, if_exists):
        """

        Args:
            func:
            progress:
            from_path: 来源路径，可能是scp.put，来源在本地，也肯是scp.get，来源在服务器
            to_dir: to_同理，跟from_相反
            to_path:
            if_exists:

        Returns:

        """

        def scp_core(mtime=False):
            a, b = self.__check_filetype(from_path, to_path)
            if b == 0:  # 目标位置不存在对应文件或目录，可以直接传输
                func(str(from_path), str(to_dir), recursive=True, preserve_times=True)
            elif a == b == 1:
                if mtime and from_path.mtime() > to_path.mtime():
                    func(str(from_path), str(to_dir), preserve_times=True)
            elif a == b == -1:
                to_paths = to_path.sub_rel_paths(mtime)
                if mtime:
                    # 因为不知道是from还是to是远程服务器，都统一一次性获得mtime更好
                    from_paths = from_path.sub_rel_paths(mtime)

                def put_dir(dir0):
                    # 1 获得当前目录文件情况
                    sub_dirs, sub_files = [], []
                    if isinstance(dir0, self.Path):
                        sub_files = [(dir0 / f) for f in dir0.sub_rel_paths(extcmd='-maxdepth 1 -type f')]
                        sub_dirs = [(dir0 / d) for d in dir0.sub_rel_paths(extcmd='-maxdepth 1 -type d')]
                    else:
                        for p in dir0.glob('*'):  # 本机文件
                            if p.is_file():
                                sub_files.append(p)
                            else:
                                sub_dirs.append(p)

                    # 2 上传文件
                    sub_files = natural_sort(sub_files)
                    for f in sub_files:
                        rf = f.relative_to(from_path).as_posix()
                        progress.kwargs['rf'] = rf
                        # 检查时间戳
                        # print(from_path, to_path, rf, from_paths[rf], to_paths.get(rf, 0),
                        #       from_paths[rf] - to_paths.get(rf, 0))
                        if rf in to_paths and ((not mtime) or (from_paths[rf] <= to_paths[rf])):
                            if progress.tqdm:
                                if isinstance(f, self.Path):
                                    # 如果不需要从服务器下载到本地，直接用本地文件尺寸代替去服务器找文件尺寸
                                    progress.tqdm.total -= (to_path / rf).size()
                                else:
                                    progress.tqdm.total -= f.size()
                        else:
                            func(str(f), str(to_path / rf), preserve_times=True)

                    # 3 上传目录
                    sub_dirs = natural_sort(sub_dirs)
                    for d in sub_dirs:
                        rd = d.relative_to(from_path).as_posix()
                        if rd in to_paths:
                            put_dir(d)
                        else:
                            func(str(d), str(to_path / rd), recursive=True, preserve_times=True)

                put_dir(from_path)

        if if_exists is None or if_exists == 'replace':
            func(str(from_path), str(to_dir), recursive=True, preserve_times=True)
        elif if_exists == 'skip':
            scp_core(False)
        elif if_exists == 'mtime':
            scp_core(True)
        else:
            raise ValueError

    def scp_get(self, remote_path=None, local_dir=None, *, local_path=None, info=True, limit_bytes=None,
                if_exists=None):
        """ 文档参考 self.scp_put

        :param local_path: 可以不输入远程remote_path，仅输入local_path来映射、运行
        get应该会比put更慢一点，因为get需要使用命令从服务器获得更多参考信息，而put很多文件信息可以在本地直接获得

        >> self.scp_get('/home/datasets/doc3D', '/home/dataset')
        """
        if local_path:
            local_path = XlPath(local_path)
            remote_path = self.__remote_dir(local_path, None) / local_path.name
            local_dir = local_path.parent
        else:
            remote_path = self.Path(remote_path)
            local_dir = self.__local_dir(remote_path, local_dir)
            local_path = local_dir / remote_path.name

        # 远程文件不存在，不用运行
        if not remote_path.exists_type():
            return

        if not local_dir.exists():
            local_dir.mkdir(parents=True, exist_ok=True)

        if info == 0:
            progress = ScpProgress(info, limit_bytes=limit_bytes)
        elif info == 1:
            progress = ScpProgress(info, desc=f'↓{local_path}',
                                   total=remote_path.size(),
                                   limit_bytes=limit_bytes)
        else:
            raise NotImplementedError

        scp = scplib.SCPClient(self.get_transport(), progress=progress)

        try:
            self.__scp_base(scp.get, progress, remote_path, local_dir, local_path, if_exists)
        except ScpLimitError:
            pass

    def scp_put(self, local_path, remote_dir=None, *, mkdir=True, print_mode=True, limit_bytes=None, if_exists=None):
        """ 将本地的local_path上传到服务器的remote_path

        差不多是标准的scp接口，可以强制上传一个文件，或者一个目录

        :param local_path: 可以是文件，也可以是目录
            上传目录时
                同名子文件会被替换
                不存在的文件会增量
                本地没有，但服务器有的文件不做操作
        :param remote_dir: 远程位置的父级目录，注意是父级
            因为DataSync侧重数据同步，所以上传的文件或目录默认同名
                如果要改名字，建议使用底层的scp接口实现
        :param mkdir: remote_dir可能不存在，为了避免出错，是否执行下mkdir
        :param bool|int print_mode:
            0: 不显示进度
            1: 显示整体上传进度（字节）
            2: 显示每个文件上传进度  （这个功能好像不常用没那么重要，暂未实现）
        :param limit_bytes: 限制传输的字节上限，可以用来筛选部分样例数据，不用下载整个庞大的目录
            注意该限制
        :param if_exists:
            None, 'replace' 不处理，也就是直接替换掉
            'skip'，跳过不处理，保留远程原文件
            'mtime'，对比时间戳，如果本地文件时间更新，则会上传到远程，否则跳过不处理

        >> self.scp_put('D:/home/chenkunze/test')
        >> self.scp_put('D:/home/chenkunze/test', '/home/chenkunze')
        """
        local_path = XlPath(local_path)
        remote_dir = self.__remote_dir(local_path, remote_dir)
        remote_path = remote_dir / local_path.name

        if mkdir:  # 判断服务器remote_dir存不存在，也要用命令，还不如直接用mkdir了
            self.exec(f'mkdir -p "{remote_dir}"')  # remote如果不存在父目录则建立

        if print_mode == 0:
            # 虽然不显示运行信息，但也要记录已上传了多少流量
            progress = ScpProgress(print_mode, limit_bytes=limit_bytes)
        elif print_mode == 1:
            progress = ScpProgress(print_mode, desc=f'↑{remote_path}',
                                   total=local_path.size(),
                                   limit_bytes=limit_bytes)
        else:
            raise NotImplementedError

        # 这里可以设置 buff_size
        scp = scplib.SCPClient(self.get_transport(), progress=progress)

        try:
            self.__scp_base(scp.put, progress, local_path, remote_dir, remote_path, if_exists)
        except ScpLimitError:
            pass

    def scp_put_brief(self, _dir):
        """ 过滤一些一般不同步的文件 """
        for p in XlPath(_dir).glob('*'):
            if p.name in ('.git', '.idea') or p.suffix == '.pyc':
                continue

            if p.is_dir():
                # 如果是目录，继续递归规则处理
                self.scp_put_brief(p)
            else:
                # 否则是没被过滤的文件，直接上传
                self.scp_put(p)

    def scp_sync(self, local_path, *, mkdir=True, info=True, limit_bytes=None):
        """ 服务器和本地目录的数据同步

        该功能目前还不稳定，还有时间戳可能无法准确复制的bug，这个bug影响
        1、无法准确对比时间戳，会冗余传送文件
        2、误判时间，导致新文件时间戳不是最新的，漏同步

        sync其实就是get、put都跑一次就行了
        注意此时limit_bytes表示的不是总流量，而是上传、下载分别的最大流量
        及没有if_exists参数，默认都通过mtime时间戳来更新
        """
        self.scp_get(local_path=local_path, info=info, limit_bytes=limit_bytes, if_exists='mtime')
        self.scp_put(local_path, mkdir=mkdir, print_mode=info, limit_bytes=limit_bytes, if_exists='mtime')

    def __3_host_trace(self):
        pass

    def set_user_passwd(self, name, passwd):
        """ 修改账号密码

        How to set user passwords using passwd without a prompt? - Ask Ubuntu:
        https://askubuntu.com/questions/80444/how-to-set-user-passwords-using-passwd-without-a-prompt
        """
        self.exec(f'usermod --password $(echo {passwd} | openssl passwd -1 -stdin) {name}')

    def add_user(self, name, passwd, sudo=False):
        """ 添加新用户

        :param name: 用户名
        :param passwd: 密码
        :param sudo: 是否开sudo权限
        """
        exists_user = self.exec("awk -F: '{ print $1}' /etc/passwd").splitlines()
        if name in exists_user:
            raise ValueError(f'{name} already exists')

        self.exec(f'useradd -d /home/{name} -s /bin/bash -m {name}')
        self.set_user_passwd(name, passwd)
        if sudo:
            self.exec(f'usermod -aG sudo {name}')

    def check_cpu_usage(self, *, print_mode=False):
        # 1 获取原始信息
        cmds = ['ps --no-headers -eo "pcpu,pmem,user"',  # 列出所有程序情况
                # 使用awk，按用户分组统计cpu（百分比）、内存总使用量（绝对量）
                """awk 'BEGIN{FS=OFS=" "}{a0[$3]+=$1; a1[$3]+=$2}END {for (i in a0) print i,a0[i],a1[i]}'""",
                'sort -rn -k1,2 -k2,3'  # 按cpu、内存使用量从大到小排序，不是必须的
                ]
        lines = self.exec('|'.join(cmds)).splitlines()

        # 2 按用户分组统计
        def f(v):
            return round(float(v), 2)

        user_usage = {}
        for line in lines:
            user, cpu, mem = line.split()
            cpu, mem = f(cpu), f(mem)
            if cpu + mem > 0.0:
                user_usage[user] = [cpu, mem]
        _total = [f(sum([v[0] for v in user_usage.values()])), f(sum([v[1] for v in user_usage.values()]))]
        user_usage['_total'] = _total

        # 3 输出
        if print_mode:
            df = pd.DataFrame.from_records([[k, v[0], v[1]] for k, v in user_usage.items()],
                                           columns=['user', '%cpu', '%mem'])
            print(df)

        del user_usage['_total']
        return user_usage

    def check_gpu_usage(self, *, print_mode=False):
        """ 检查(每个用户)显存使用量

        使用这个命令，必须要求远程服务器安装了pip install gpustat

        TODO 加上记录最高显卡温度？
        """
        # 1 获取原始信息
        lines = self.exec('gpustat').splitlines()[1:]
        if print_mode:
            print('\n'.join(lines))

        # 2 按用户分组统计
        user_usage = defaultdict(float)
        for line in lines:
            name, temperature, capacity, uses = line.split('|')
            used, total = map(int, re.findall(r'\d+', capacity))
            user_usage['_other'] += used / 1024  # 会有些没有用户操作的僵尸显存占用
            for x in uses.split():
                user, one_used = re.search(r'(.+?)\((\d+)M\)', x).groups()
                one_used = int(one_used) / 1024
                user_usage[user] += one_used
                user_usage['_other'] -= one_used

        # 使用量从多到少排序。但注意如果转存到PG，jsonb会重新排序。
        user_usage = {k: round(user_usage[k], 2) for k in sorted(user_usage, key=lambda k: -user_usage[k])}
        if user_usage['_other'] < 0.01:
            del user_usage['_other']

        user_usage['_total'] = round(sum(user_usage.values()), 2)

        # 3 输出
        if print_mode:
            df = pd.DataFrame.from_records([[k, v] for k, v in user_usage.items()], columns=['user', 'gpu_mem_GB'])
            print(df)

        del user_usage['_total']
        return user_usage

    def check_disk_usage(self, *, print_mode=False, timeout=1200):
        """ 检查(每个用户)磁盘空间使用量

        :return: total 总字节MBytes数, msg 所有用户、包括其他非/home目录的使用MBytes数
        """
        GB = 1024 ** 2  # df、du本身默认单位已经是KB，所以2次方后，就是GB了
        # 1 整体情况
        used, total_memory = 0, 0
        for line in self.exec('df').splitlines()[1:]:
            if not line.startswith('/dev/'):
                continue
            # 为了避免遇到路径中有空格的问题，用正则做了较复杂的判断、切割
            _total, _used = map(int, re.search(r'\s+(\d+)\s+(\d+)\s+\d+\s+\d+', line).groups())
            used += _used
            total_memory += _total

        # 2 /home目录下每个占用情况
        user_usage = defaultdict(int)
        for line in self.exec('du -d 1 /home', timeout=timeout).splitlines():  # 这个要限时，默认20分钟
            _bytes, _dir = line.split(maxsplit=1)
            _dir = _dir[6:]
            if _dir and int(_bytes) > GB:  # 达到1GB的才记录
                user_usage[_dir] += int(_bytes)

        user_usage['_other'] = used - sum(user_usage.values())
        user_usage = {k: (v // GB) for k, v in user_usage.items()}
        user_usage = {k: round(user_usage[k], 2) for k in sorted(user_usage, key=lambda k: -user_usage[k])}
        user_usage['_total'] = total_memory // GB

        # 3 展示
        if print_mode:
            df = pd.DataFrame.from_records([[k, v] for k, v in user_usage.items()], columns=['user', 'disk_mem_GB'])
            print(df)

        del user_usage['_total']
        return user_usage

    def __4_其他(self):
        pass

    def restart_frpc(self, frp_dir='/root/frp_0.37.0_linux_amd64'):
        """ 因为service不一定有效，这里通过暴力找frpc的方式来设置

        :param frp_dir: frp文件所在目录
        """
        cmds = [
            # 找到已有的 ./frpc 关闭
            r"for i in $(ps -eo 'args,pid' | awk '/^.\/frpc / {print $2}'); do kill ${i}; done",
            # 重新启动 frpc
            f"cd {frp_dir}; nohup ./frpc > /dev/null 2>&1 &"
        ]
        # 注意关闭和重启需要同时操作，不然在外网穿刺连接ssh，执行第1句后就断开连接了
        self.exec('; '.join(cmds))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/08/20 11:46


import re


def py_remove_interaction_chars(s):
    """ 去掉复制的一段代码中，前导的“>>>”标记 """
    # 这个算法可能还不够严谨，实际应用中再逐步写鲁棒
    # ">>> "、"... "
    lines = [line[4:] for line in s.splitlines()]
    return '\n'.join(lines)


def pycode_sort_import(s):
    from pyxllib.text.nestenv import PyNestEnv

    def cmp(line):
        """ 将任意一句import映射为一个可比较的list对象

        :return: 2个数值
            1、模块优先级
            2、import在前，from在后
        """
        name = re.search(r'(?:import|from)\s+(\S+)', line).group(1)
        for i, x in enumerate('stdlib prog algo text file debug cv data database gui ai robot tool ex'.split()):
            name = name.replace('pyxllib.' + x, f'{i:02}')
        for i, x in enumerate('pyxllib pyxlpr xlproject'.split()):
            name = name.replace(x, f'~{i:02}')
        for i, x in enumerate('newbie pupil specialist expert'.split()):
            name = name.replace('.' + x, f'{i:02}')

        # 忽略大小写
        return [name.lower(), line.startswith('import')]

    def sort_part(m):
        parts = PyNestEnv(m.group()).imports().strings()
        parts = [p.rstrip() + '\n' for p in parts]
        parts.sort(key=cmp)
        return ''.join(parts)

    res = PyNestEnv(s).imports().sub(sort_part, adjacent=True)  # 需要邻接，分块处理
    return res



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["build_backbone"]


def build_backbone(config, model_type):
    if model_type == "det" or model_type == "table":
        from .det_mobilenet_v3 import MobileNetV3
        from .det_resnet_vd import ResNet
        from .det_resnet_vd_sast import ResNet_SAST
        support_dict = ["MobileNetV3", "ResNet", "ResNet_SAST"]
    elif model_type == "rec" or model_type == "cls":
        from .rec_mobilenet_v3 import MobileNetV3
        from .rec_resnet_vd import ResNet
        from .rec_resnet_fpn import ResNetFPN
        from .rec_mv1_enhance import MobileNetV1Enhance
        from .rec_nrtr_mtb import MTB
        from .rec_resnet_31 import ResNet31
        from .rec_resnet_aster import ResNet_ASTER
        support_dict = [
            'MobileNetV1Enhance', 'MobileNetV3', 'ResNet', 'ResNetFPN', 'MTB',
            "ResNet31", "ResNet_ASTER"
        ]
    elif model_type == "e2e":
        from .e2e_resnet_vd_pg import ResNet
        support_dict = ['ResNet']
    elif model_type == 'kie':
        from .kie_unet_sdmgr import Kie_backbone
        support_dict = ['Kie_backbone']
    elif model_type == "table":
        from .table_resnet_vd import ResNet
        from .table_mobilenet_v3 import MobileNetV3
        support_dict = ["ResNet", "MobileNetV3"]
    else:
        raise NotImplementedError

    module_name = config.pop("name")
    assert module_name in support_dict, Exception(
        "when model typs is {}, backbone only support {}".format(model_type,
                                                                 support_dict))
    module_class = eval(module_name)(**config)
    return module_class



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is refer from: https://github.com/PaddlePaddle/PaddleClas/blob/develop/ppcls/arch/backbone/legendary_models/pp_lcnet.py

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import paddle
from paddle import ParamAttr
import paddle.nn as nn
import paddle.nn.functional as F
from paddle.nn import Conv2D, BatchNorm, Linear, Dropout
from paddle.nn import AdaptiveAvgPool2D, MaxPool2D, AvgPool2D
from paddle.nn.initializer import KaimingNormal
import math
import numpy as np
import paddle
from paddle import ParamAttr, reshape, transpose, concat, split
import paddle.nn as nn
import paddle.nn.functional as F
from paddle.nn import Conv2D, BatchNorm, Linear, Dropout
from paddle.nn import AdaptiveAvgPool2D, MaxPool2D, AvgPool2D
from paddle.nn.initializer import KaimingNormal
import math
from paddle.nn.functional import hardswish, hardsigmoid
from paddle.regularizer import L2Decay


class ConvBNLayer(nn.Layer):
    def __init__(self,
                 num_channels,
                 filter_size,
                 num_filters,
                 stride,
                 padding,
                 channels=None,
                 num_groups=1,
                 act='hard_swish'):
        super(ConvBNLayer, self).__init__()

        self._conv = Conv2D(
            in_channels=num_channels,
            out_channels=num_filters,
            kernel_size=filter_size,
            stride=stride,
            padding=padding,
            groups=num_groups,
            weight_attr=ParamAttr(initializer=KaimingNormal()),
            bias_attr=False)

        self._batch_norm = BatchNorm(
            num_filters,
            act=act,
            param_attr=ParamAttr(regularizer=L2Decay(0.0)),
            bias_attr=ParamAttr(regularizer=L2Decay(0.0)))

    def forward(self, inputs):
        y = self._conv(inputs)
        y = self._batch_norm(y)
        return y


class DepthwiseSeparable(nn.Layer):
    def __init__(self,
                 num_channels,
                 num_filters1,
                 num_filters2,
                 num_groups,
                 stride,
                 scale,
                 dw_size=3,
                 padding=1,
                 use_se=False):
        super(DepthwiseSeparable, self).__init__()
        self.use_se = use_se
        self._depthwise_conv = ConvBNLayer(
            num_channels=num_channels,
            num_filters=int(num_filters1 * scale),
            filter_size=dw_size,
            stride=stride,
            padding=padding,
            num_groups=int(num_groups * scale))
        if use_se:
            self._se = SEModule(int(num_filters1 * scale))
        self._pointwise_conv = ConvBNLayer(
            num_channels=int(num_filters1 * scale),
            filter_size=1,
            num_filters=int(num_filters2 * scale),
            stride=1,
            padding=0)

    def forward(self, inputs):
        y = self._depthwise_conv(inputs)
        if self.use_se:
            y = self._se(y)
        y = self._pointwise_conv(y)
        return y


class MobileNetV1Enhance(nn.Layer):
    def __init__(self, in_channels=3, scale=0.5, **kwargs):
        super().__init__()
        self.scale = scale
        self.block_list = []

        self.conv1 = ConvBNLayer(
            num_channels=3,
            filter_size=3,
            channels=3,
            num_filters=int(32 * scale),
            stride=2,
            padding=1)

        conv2_1 = DepthwiseSeparable(
            num_channels=int(32 * scale),
            num_filters1=32,
            num_filters2=64,
            num_groups=32,
            stride=1,
            scale=scale)
        self.block_list.append(conv2_1)

        conv2_2 = DepthwiseSeparable(
            num_channels=int(64 * scale),
            num_filters1=64,
            num_filters2=128,
            num_groups=64,
            stride=1,
            scale=scale)
        self.block_list.append(conv2_2)

        conv3_1 = DepthwiseSeparable(
            num_channels=int(128 * scale),
            num_filters1=128,
            num_filters2=128,
            num_groups=128,
            stride=1,
            scale=scale)
        self.block_list.append(conv3_1)

        conv3_2 = DepthwiseSeparable(
            num_channels=int(128 * scale),
            num_filters1=128,
            num_filters2=256,
            num_groups=128,
            stride=(2, 1),
            scale=scale)
        self.block_list.append(conv3_2)

        conv4_1 = DepthwiseSeparable(
            num_channels=int(256 * scale),
            num_filters1=256,
            num_filters2=256,
            num_groups=256,
            stride=1,
            scale=scale)
        self.block_list.append(conv4_1)

        conv4_2 = DepthwiseSeparable(
            num_channels=int(256 * scale),
            num_filters1=256,
            num_filters2=512,
            num_groups=256,
            stride=(2, 1),
            scale=scale)
        self.block_list.append(conv4_2)

        for _ in range(5):
            conv5 = DepthwiseSeparable(
                num_channels=int(512 * scale),
                num_filters1=512,
                num_filters2=512,
                num_groups=512,
                stride=1,
                dw_size=5,
                padding=2,
                scale=scale,
                use_se=False)
            self.block_list.append(conv5)

        conv5_6 = DepthwiseSeparable(
            num_channels=int(512 * scale),
            num_filters1=512,
            num_filters2=1024,
            num_groups=512,
            stride=(2, 1),
            dw_size=5,
            padding=2,
            scale=scale,
            use_se=True)
        self.block_list.append(conv5_6)

        conv6 = DepthwiseSeparable(
            num_channels=int(1024 * scale),
            num_filters1=1024,
            num_filters2=1024,
            num_groups=1024,
            stride=1,
            dw_size=5,
            padding=2,
            use_se=True,
            scale=scale)
        self.block_list.append(conv6)

        self.block_list = nn.Sequential(*self.block_list)

        self.pool = nn.MaxPool2D(kernel_size=2, stride=2, padding=0)
        self.out_channels = int(1024 * scale)

    def forward(self, inputs):
        y = self.conv1(inputs)
        y = self.block_list(y)
        y = self.pool(y)
        return y


class SEModule(nn.Layer):
    def __init__(self, channel, reduction=4):
        super(SEModule, self).__init__()
        self.avg_pool = AdaptiveAvgPool2D(1)
        self.conv1 = Conv2D(
            in_channels=channel,
            out_channels=channel // reduction,
            kernel_size=1,
            stride=1,
            padding=0,
            weight_attr=ParamAttr(),
            bias_attr=ParamAttr())
        self.conv2 = Conv2D(
            in_channels=channel // reduction,
            out_channels=channel,
            kernel_size=1,
            stride=1,
            padding=0,
            weight_attr=ParamAttr(),
            bias_attr=ParamAttr())

    def forward(self, inputs):
        outputs = self.avg_pool(inputs)
        outputs = self.conv1(outputs)
        outputs = F.relu(outputs)
        outputs = self.conv2(outputs)
        outputs = hardsigmoid(outputs)
        return paddle.multiply(x=inputs, y=outputs)



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import paddle
import copy
from paddle import nn
import paddle.nn.functional as F
from paddle.nn import LayerList
from paddle.nn.initializer import XavierNormal as xavier_uniform_
from paddle.nn import Dropout, Linear, LayerNorm, Conv2D
import numpy as np
from pyxlpr.ppocr.modeling.heads.multiheadAttention import MultiheadAttention
from paddle.nn.initializer import Constant as constant_
from paddle.nn.initializer import XavierNormal as xavier_normal_

zeros_ = constant_(value=0.)
ones_ = constant_(value=1.)


class Transformer(nn.Layer):
    """A transformer model. User is able to modify the attributes as needed. The architechture
    is based on the paper "Attention Is All You Need". Ashish Vaswani, Noam Shazeer,
    Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez, Lukasz Kaiser, and
    Illia Polosukhin. 2017. Attention is all you need. In Advances in Neural Information
    Processing Systems, pages 6000-6010.

    Args:
        d_model: the number of expected features in the encoder/decoder inputs (default=512).
        nhead: the number of heads in the multiheadattention models (default=8).
        num_encoder_layers: the number of sub-encoder-layers in the encoder (default=6).
        num_decoder_layers: the number of sub-decoder-layers in the decoder (default=6).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        custom_encoder: custom encoder (default=None).
        custom_decoder: custom decoder (default=None).

    """

    def __init__(self,
                 d_model=512,
                 nhead=8,
                 num_encoder_layers=6,
                 beam_size=0,
                 num_decoder_layers=6,
                 dim_feedforward=1024,
                 attention_dropout_rate=0.0,
                 residual_dropout_rate=0.1,
                 custom_encoder=None,
                 custom_decoder=None,
                 in_channels=0,
                 out_channels=0,
                 scale_embedding=True):
        super(Transformer, self).__init__()
        self.out_channels = out_channels + 1
        self.embedding = Embeddings(
            d_model=d_model,
            vocab=self.out_channels,
            padding_idx=0,
            scale_embedding=scale_embedding)
        self.positional_encoding = PositionalEncoding(
            dropout=residual_dropout_rate,
            dim=d_model, )
        if custom_encoder is not None:
            self.encoder = custom_encoder
        else:
            if num_encoder_layers > 0:
                encoder_layer = TransformerEncoderLayer(
                    d_model, nhead, dim_feedforward, attention_dropout_rate,
                    residual_dropout_rate)
                self.encoder = TransformerEncoder(encoder_layer,
                                                  num_encoder_layers)
            else:
                self.encoder = None

        if custom_decoder is not None:
            self.decoder = custom_decoder
        else:
            decoder_layer = TransformerDecoderLayer(
                d_model, nhead, dim_feedforward, attention_dropout_rate,
                residual_dropout_rate)
            self.decoder = TransformerDecoder(decoder_layer, num_decoder_layers)

        self._reset_parameters()
        self.beam_size = beam_size
        self.d_model = d_model
        self.nhead = nhead
        self.tgt_word_prj = nn.Linear(
            d_model, self.out_channels, bias_attr=False)
        w0 = np.random.normal(0.0, d_model**-0.5,
                              (d_model, self.out_channels)).astype(np.float32)
        self.tgt_word_prj.weight.set_value(w0)
        self.apply(self._init_weights)

    def _init_weights(self, m):

        if isinstance(m, nn.Conv2D):
            xavier_normal_(m.weight)
            if m.bias is not None:
                zeros_(m.bias)

    def forward_train(self, src, tgt):
        tgt = tgt[:, :-1]

        tgt_key_padding_mask = self.generate_padding_mask(tgt)
        tgt = self.embedding(tgt).transpose([1, 0, 2])
        tgt = self.positional_encoding(tgt)
        tgt_mask = self.generate_square_subsequent_mask(tgt.shape[0])

        if self.encoder is not None:
            src = self.positional_encoding(src.transpose([1, 0, 2]))
            memory = self.encoder(src)
        else:
            memory = src.squeeze(2).transpose([2, 0, 1])
        output = self.decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            memory_mask=None,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=None)
        output = output.transpose([1, 0, 2])
        logit = self.tgt_word_prj(output)
        return logit

    def forward(self, src, targets=None):
        """Take in and process masked source/target sequences.
        Args:
            src: the sequence to the encoder (required).
            tgt: the sequence to the decoder (required).
        Shape:
            - src: :math:`(S, N, E)`.
            - tgt: :math:`(T, N, E)`.
        Examples:
            >>> output = transformer_model(src, tgt)
        """

        if self.training:
            max_len = targets[1].max()
            tgt = targets[0][:, :2 + max_len]
            return self.forward_train(src, tgt)
        else:
            if self.beam_size > 0:
                return self.forward_beam(src)
            else:
                return self.forward_test(src)

    def forward_test(self, src):
        bs = paddle.shape(src)[0]
        if self.encoder is not None:
            src = self.positional_encoding(paddle.transpose(src, [1, 0, 2]))
            memory = self.encoder(src)
        else:
            memory = paddle.transpose(paddle.squeeze(src, 2), [2, 0, 1])
        dec_seq = paddle.full((bs, 1), 2, dtype=paddle.int64)
        dec_prob = paddle.full((bs, 1), 1., dtype=paddle.float32)
        for len_dec_seq in range(1, 25):
            dec_seq_embed = paddle.transpose(self.embedding(dec_seq), [1, 0, 2])
            dec_seq_embed = self.positional_encoding(dec_seq_embed)
            tgt_mask = self.generate_square_subsequent_mask(
                paddle.shape(dec_seq_embed)[0])
            output = self.decoder(
                dec_seq_embed,
                memory,
                tgt_mask=tgt_mask,
                memory_mask=None,
                tgt_key_padding_mask=None,
                memory_key_padding_mask=None)
            dec_output = paddle.transpose(output, [1, 0, 2])
            dec_output = dec_output[:, -1, :]
            word_prob = F.softmax(self.tgt_word_prj(dec_output), axis=1)
            preds_idx = paddle.argmax(word_prob, axis=1)
            if paddle.equal_all(
                    preds_idx,
                    paddle.full(
                        paddle.shape(preds_idx), 3, dtype='int64')):
                break
            preds_prob = paddle.max(word_prob, axis=1)
            dec_seq = paddle.concat(
                [dec_seq, paddle.reshape(preds_idx, [-1, 1])], axis=1)
            dec_prob = paddle.concat(
                [dec_prob, paddle.reshape(preds_prob, [-1, 1])], axis=1)
        return [dec_seq, dec_prob]

    def forward_beam(self, images):
        ''' Translation work in one batch '''

        def get_inst_idx_to_tensor_position_map(inst_idx_list):
            ''' Indicate the position of an instance in a tensor. '''
            return {
                inst_idx: tensor_position
                for tensor_position, inst_idx in enumerate(inst_idx_list)
            }

        def collect_active_part(beamed_tensor, curr_active_inst_idx,
                                n_prev_active_inst, n_bm):
            ''' Collect tensor parts associated to active instances. '''

            beamed_tensor_shape = paddle.shape(beamed_tensor)
            n_curr_active_inst = len(curr_active_inst_idx)
            new_shape = (n_curr_active_inst * n_bm, beamed_tensor_shape[1],
                         beamed_tensor_shape[2])

            beamed_tensor = beamed_tensor.reshape([n_prev_active_inst, -1])
            beamed_tensor = beamed_tensor.index_select(
                curr_active_inst_idx, axis=0)
            beamed_tensor = beamed_tensor.reshape(new_shape)

            return beamed_tensor

        def collate_active_info(src_enc, inst_idx_to_position_map,
                                active_inst_idx_list):
            # Sentences which are still active are collected,
            # so the decoder will not run on completed sentences.

            n_prev_active_inst = len(inst_idx_to_position_map)
            active_inst_idx = [
                inst_idx_to_position_map[k] for k in active_inst_idx_list
            ]
            active_inst_idx = paddle.to_tensor(active_inst_idx, dtype='int64')
            active_src_enc = collect_active_part(
                src_enc.transpose([1, 0, 2]), active_inst_idx,
                n_prev_active_inst, n_bm).transpose([1, 0, 2])
            active_inst_idx_to_position_map = get_inst_idx_to_tensor_position_map(
                active_inst_idx_list)
            return active_src_enc, active_inst_idx_to_position_map

        def beam_decode_step(inst_dec_beams, len_dec_seq, enc_output,
                             inst_idx_to_position_map, n_bm,
                             memory_key_padding_mask):
            ''' Decode and update beam status, and then return active beam idx '''

            def prepare_beam_dec_seq(inst_dec_beams, len_dec_seq):
                dec_partial_seq = [
                    b.get_current_state() for b in inst_dec_beams if not b.done
                ]
                dec_partial_seq = paddle.stack(dec_partial_seq)
                dec_partial_seq = dec_partial_seq.reshape([-1, len_dec_seq])
                return dec_partial_seq

            def predict_word(dec_seq, enc_output, n_active_inst, n_bm,
                             memory_key_padding_mask):
                dec_seq = paddle.transpose(self.embedding(dec_seq), [1, 0, 2])
                dec_seq = self.positional_encoding(dec_seq)
                tgt_mask = self.generate_square_subsequent_mask(
                    paddle.shape(dec_seq)[0])
                dec_output = self.decoder(
                    dec_seq,
                    enc_output,
                    tgt_mask=tgt_mask,
                    tgt_key_padding_mask=None,
                    memory_key_padding_mask=memory_key_padding_mask, )
                dec_output = paddle.transpose(dec_output, [1, 0, 2])
                dec_output = dec_output[:,
                                        -1, :]  # Pick the last step: (bh * bm) * d_h
                word_prob = F.softmax(self.tgt_word_prj(dec_output), axis=1)
                word_prob = paddle.reshape(word_prob, [n_active_inst, n_bm, -1])
                return word_prob

            def collect_active_inst_idx_list(inst_beams, word_prob,
                                             inst_idx_to_position_map):
                active_inst_idx_list = []
                for inst_idx, inst_position in inst_idx_to_position_map.items():
                    is_inst_complete = inst_beams[inst_idx].advance(word_prob[
                        inst_position])
                    if not is_inst_complete:
                        active_inst_idx_list += [inst_idx]

                return active_inst_idx_list

            n_active_inst = len(inst_idx_to_position_map)
            dec_seq = prepare_beam_dec_seq(inst_dec_beams, len_dec_seq)
            word_prob = predict_word(dec_seq, enc_output, n_active_inst, n_bm,
                                     None)
            # Update the beam with predicted word prob information and collect incomplete instances
            active_inst_idx_list = collect_active_inst_idx_list(
                inst_dec_beams, word_prob, inst_idx_to_position_map)
            return active_inst_idx_list

        def collect_hypothesis_and_scores(inst_dec_beams, n_best):
            all_hyp, all_scores = [], []
            for inst_idx in range(len(inst_dec_beams)):
                scores, tail_idxs = inst_dec_beams[inst_idx].sort_scores()
                all_scores += [scores[:n_best]]
                hyps = [
                    inst_dec_beams[inst_idx].get_hypothesis(i)
                    for i in tail_idxs[:n_best]
                ]
                all_hyp += [hyps]
            return all_hyp, all_scores

        with paddle.no_grad():
            #-- Encode
            if self.encoder is not None:
                src = self.positional_encoding(images.transpose([1, 0, 2]))
                src_enc = self.encoder(src)
            else:
                src_enc = images.squeeze(2).transpose([0, 2, 1])

            n_bm = self.beam_size
            src_shape = paddle.shape(src_enc)
            inst_dec_beams = [Beam(n_bm) for _ in range(1)]
            active_inst_idx_list = list(range(1))
            # Repeat data for beam search
            src_enc = paddle.tile(src_enc, [1, n_bm, 1])
            inst_idx_to_position_map = get_inst_idx_to_tensor_position_map(
                active_inst_idx_list)
            # Decode
            for len_dec_seq in range(1, 25):
                src_enc_copy = src_enc.clone()
                active_inst_idx_list = beam_decode_step(
                    inst_dec_beams, len_dec_seq, src_enc_copy,
                    inst_idx_to_position_map, n_bm, None)
                if not active_inst_idx_list:
                    break  # all instances have finished their path to <EOS>
                src_enc, inst_idx_to_position_map = collate_active_info(
                    src_enc_copy, inst_idx_to_position_map,
                    active_inst_idx_list)
        batch_hyp, batch_scores = collect_hypothesis_and_scores(inst_dec_beams,
                                                                1)
        result_hyp = []
        hyp_scores = []
        for bs_hyp, score in zip(batch_hyp, batch_scores):
            l = len(bs_hyp[0])
            bs_hyp_pad = bs_hyp[0] + [3] * (25 - l)
            result_hyp.append(bs_hyp_pad)
            score = float(score) / l
            hyp_score = [score for _ in range(25)]
            hyp_scores.append(hyp_score)
        return [
            paddle.to_tensor(
                np.array(result_hyp), dtype=paddle.int64),
            paddle.to_tensor(hyp_scores)
        ]

    def generate_square_subsequent_mask(self, sz):
        """Generate a square mask for the sequence. The masked positions are filled with float('-inf').
            Unmasked positions are filled with float(0.0).
        """
        mask = paddle.zeros([sz, sz], dtype='float32')
        mask_inf = paddle.triu(
            paddle.full(
                shape=[sz, sz], dtype='float32', fill_value='-inf'),
            diagonal=1)
        mask = mask + mask_inf
        return mask

    def generate_padding_mask(self, x):
        padding_mask = paddle.equal(x, paddle.to_tensor(0, dtype=x.dtype))
        return padding_mask

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""

        for p in self.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)


class TransformerEncoder(nn.Layer):
    """TransformerEncoder is a stack of N encoder layers
    Args:
        encoder_layer: an instance of the TransformerEncoderLayer() class (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).
    """

    def __init__(self, encoder_layer, num_layers):
        super(TransformerEncoder, self).__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers

    def forward(self, src):
        """Pass the input through the endocder layers in turn.
        Args:
            src: the sequnce to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
        """
        output = src

        for i in range(self.num_layers):
            output = self.layers[i](output,
                                    src_mask=None,
                                    src_key_padding_mask=None)

        return output


class TransformerDecoder(nn.Layer):
    """TransformerDecoder is a stack of N decoder layers

    Args:
        decoder_layer: an instance of the TransformerDecoderLayer() class (required).
        num_layers: the number of sub-decoder-layers in the decoder (required).
        norm: the layer normalization component (optional).

    """

    def __init__(self, decoder_layer, num_layers):
        super(TransformerDecoder, self).__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers

    def forward(self,
                tgt,
                memory,
                tgt_mask=None,
                memory_mask=None,
                tgt_key_padding_mask=None,
                memory_key_padding_mask=None):
        """Pass the inputs (and mask) through the decoder layer in turn.

        Args:
            tgt: the sequence to the decoder (required).
            memory: the sequnce from the last layer of the encoder (required).
            tgt_mask: the mask for the tgt sequence (optional).
            memory_mask: the mask for the memory sequence (optional).
            tgt_key_padding_mask: the mask for the tgt keys per batch (optional).
            memory_key_padding_mask: the mask for the memory keys per batch (optional).
        """
        output = tgt
        for i in range(self.num_layers):
            output = self.layers[i](
                output,
                memory,
                tgt_mask=tgt_mask,
                memory_mask=memory_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                memory_key_padding_mask=memory_key_padding_mask)

        return output


class TransformerEncoderLayer(nn.Layer):
    """TransformerEncoderLayer is made up of self-attn and feedforward network.
    This standard encoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).

    """

    def __init__(self,
                 d_model,
                 nhead,
                 dim_feedforward=2048,
                 attention_dropout_rate=0.0,
                 residual_dropout_rate=0.1):
        super(TransformerEncoderLayer, self).__init__()
        self.self_attn = MultiheadAttention(
            d_model, nhead, dropout=attention_dropout_rate)

        self.conv1 = Conv2D(
            in_channels=d_model,
            out_channels=dim_feedforward,
            kernel_size=(1, 1))
        self.conv2 = Conv2D(
            in_channels=dim_feedforward,
            out_channels=d_model,
            kernel_size=(1, 1))

        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.dropout1 = Dropout(residual_dropout_rate)
        self.dropout2 = Dropout(residual_dropout_rate)

    def forward(self, src, src_mask=None, src_key_padding_mask=None):
        """Pass the input through the endocder layer.
        Args:
            src: the sequnce to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
        """
        src2 = self.self_attn(
            src,
            src,
            src,
            attn_mask=src_mask,
            key_padding_mask=src_key_padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        src = paddle.transpose(src, [1, 2, 0])
        src = paddle.unsqueeze(src, 2)
        src2 = self.conv2(F.relu(self.conv1(src)))
        src2 = paddle.squeeze(src2, 2)
        src2 = paddle.transpose(src2, [2, 0, 1])
        src = paddle.squeeze(src, 2)
        src = paddle.transpose(src, [2, 0, 1])

        src = src + self.dropout2(src2)
        src = self.norm2(src)
        return src


class TransformerDecoderLayer(nn.Layer):
    """TransformerDecoderLayer is made up of self-attn, multi-head-attn and feedforward network.
    This standard decoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).

    """

    def __init__(self,
                 d_model,
                 nhead,
                 dim_feedforward=2048,
                 attention_dropout_rate=0.0,
                 residual_dropout_rate=0.1):
        super(TransformerDecoderLayer, self).__init__()
        self.self_attn = MultiheadAttention(
            d_model, nhead, dropout=attention_dropout_rate)
        self.multihead_attn = MultiheadAttention(
            d_model, nhead, dropout=attention_dropout_rate)

        self.conv1 = Conv2D(
            in_channels=d_model,
            out_channels=dim_feedforward,
            kernel_size=(1, 1))
        self.conv2 = Conv2D(
            in_channels=dim_feedforward,
            out_channels=d_model,
            kernel_size=(1, 1))

        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.norm3 = LayerNorm(d_model)
        self.dropout1 = Dropout(residual_dropout_rate)
        self.dropout2 = Dropout(residual_dropout_rate)
        self.dropout3 = Dropout(residual_dropout_rate)

    def forward(self,
                tgt,
                memory,
                tgt_mask=None,
                memory_mask=None,
                tgt_key_padding_mask=None,
                memory_key_padding_mask=None):
        """Pass the inputs (and mask) through the decoder layer.

        Args:
            tgt: the sequence to the decoder layer (required).
            memory: the sequnce from the last layer of the encoder (required).
            tgt_mask: the mask for the tgt sequence (optional).
            memory_mask: the mask for the memory sequence (optional).
            tgt_key_padding_mask: the mask for the tgt keys per batch (optional).
            memory_key_padding_mask: the mask for the memory keys per batch (optional).

        """
        tgt2 = self.self_attn(
            tgt,
            tgt,
            tgt,
            attn_mask=tgt_mask,
            key_padding_mask=tgt_key_padding_mask)
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)
        tgt2 = self.multihead_attn(
            tgt,
            memory,
            memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask)
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)

        # default
        tgt = paddle.transpose(tgt, [1, 2, 0])
        tgt = paddle.unsqueeze(tgt, 2)
        tgt2 = self.conv2(F.relu(self.conv1(tgt)))
        tgt2 = paddle.squeeze(tgt2, 2)
        tgt2 = paddle.transpose(tgt2, [2, 0, 1])
        tgt = paddle.squeeze(tgt, 2)
        tgt = paddle.transpose(tgt, [2, 0, 1])

        tgt = tgt + self.dropout3(tgt2)
        tgt = self.norm3(tgt)
        return tgt


def _get_clones(module, N):
    return LayerList([copy.deepcopy(module) for i in range(N)])


class PositionalEncoding(nn.Layer):
    """Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, dropout, dim, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = paddle.zeros([max_len, dim])
        position = paddle.arange(0, max_len, dtype=paddle.float32).unsqueeze(1)
        div_term = paddle.exp(
            paddle.arange(0, dim, 2).astype('float32') *
            (-math.log(10000.0) / dim))
        pe[:, 0::2] = paddle.sin(position * div_term)
        pe[:, 1::2] = paddle.cos(position * div_term)
        pe = paddle.unsqueeze(pe, 0)
        pe = paddle.transpose(pe, [1, 0, 2])
        self.register_buffer('pe', pe)

    def forward(self, x):
        """Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """
        x = x + self.pe[:paddle.shape(x)[0], :]
        return self.dropout(x)


class PositionalEncoding_2d(nn.Layer):
    """Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, dropout, dim, max_len=5000):
        super(PositionalEncoding_2d, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = paddle.zeros([max_len, dim])
        position = paddle.arange(0, max_len, dtype=paddle.float32).unsqueeze(1)
        div_term = paddle.exp(
            paddle.arange(0, dim, 2).astype('float32') *
            (-math.log(10000.0) / dim))
        pe[:, 0::2] = paddle.sin(position * div_term)
        pe[:, 1::2] = paddle.cos(position * div_term)
        pe = paddle.transpose(paddle.unsqueeze(pe, 0), [1, 0, 2])
        self.register_buffer('pe', pe)

        self.avg_pool_1 = nn.AdaptiveAvgPool2D((1, 1))
        self.linear1 = nn.Linear(dim, dim)
        self.linear1.weight.data.fill_(1.)
        self.avg_pool_2 = nn.AdaptiveAvgPool2D((1, 1))
        self.linear2 = nn.Linear(dim, dim)
        self.linear2.weight.data.fill_(1.)

    def forward(self, x):
        """Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """
        w_pe = self.pe[:paddle.shape(x)[-1], :]
        w1 = self.linear1(self.avg_pool_1(x).squeeze()).unsqueeze(0)
        w_pe = w_pe * w1
        w_pe = paddle.transpose(w_pe, [1, 2, 0])
        w_pe = paddle.unsqueeze(w_pe, 2)

        h_pe = self.pe[:paddle.shape(x).shape[-2], :]
        w2 = self.linear2(self.avg_pool_2(x).squeeze()).unsqueeze(0)
        h_pe = h_pe * w2
        h_pe = paddle.transpose(h_pe, [1, 2, 0])
        h_pe = paddle.unsqueeze(h_pe, 3)

        x = x + w_pe + h_pe
        x = paddle.transpose(
            paddle.reshape(x,
                           [x.shape[0], x.shape[1], x.shape[2] * x.shape[3]]),
            [2, 0, 1])

        return self.dropout(x)


class Embeddings(nn.Layer):
    def __init__(self, d_model, vocab, padding_idx, scale_embedding):
        super(Embeddings, self).__init__()
        self.embedding = nn.Embedding(vocab, d_model, padding_idx=padding_idx)
        w0 = np.random.normal(0.0, d_model**-0.5,
                              (vocab, d_model)).astype(np.float32)
        self.embedding.weight.set_value(w0)
        self.d_model = d_model
        self.scale_embedding = scale_embedding

    def forward(self, x):
        if self.scale_embedding:
            x = self.embedding(x)
            return x * math.sqrt(self.d_model)
        return self.embedding(x)


class Beam():
    ''' Beam search '''

    def __init__(self, size, device=False):

        self.size = size
        self._done = False
        # The score for each translation on the beam.
        self.scores = paddle.zeros((size, ), dtype=paddle.float32)
        self.all_scores = []
        # The backpointers at each time-step.
        self.prev_ks = []
        # The outputs at each time-step.
        self.next_ys = [paddle.full((size, ), 0, dtype=paddle.int64)]
        self.next_ys[0][0] = 2

    def get_current_state(self):
        "Get the outputs for the current timestep."
        return self.get_tentative_hypothesis()

    def get_current_origin(self):
        "Get the backpointers for the current timestep."
        return self.prev_ks[-1]

    @property
    def done(self):
        return self._done

    def advance(self, word_prob):
        "Update beam status and check if finished or not."
        num_words = word_prob.shape[1]

        # Sum the previous scores.
        if len(self.prev_ks) > 0:
            beam_lk = word_prob + self.scores.unsqueeze(1).expand_as(word_prob)
        else:
            beam_lk = word_prob[0]

        flat_beam_lk = beam_lk.reshape([-1])
        best_scores, best_scores_id = flat_beam_lk.topk(self.size, 0, True,
                                                        True)  # 1st sort
        self.all_scores.append(self.scores)
        self.scores = best_scores
        # bestScoresId is flattened as a (beam x word) array,
        # so we need to calculate which word and beam each score came from
        prev_k = best_scores_id // num_words
        self.prev_ks.append(prev_k)
        self.next_ys.append(best_scores_id - prev_k * num_words)
        # End condition is when top-of-beam is EOS.
        if self.next_ys[-1][0] == 3:
            self._done = True
            self.all_scores.append(self.scores)

        return self._done

    def sort_scores(self):
        "Sort the scores."
        return self.scores, paddle.to_tensor(
            [i for i in range(int(self.scores.shape[0]))], dtype='int32')

    def get_the_best_score_and_idx(self):
        "Get the score of the best in the beam."
        scores, ids = self.sort_scores()
        return scores[1], ids[1]

    def get_tentative_hypothesis(self):
        "Get the decoded sequence for the current timestep."
        if len(self.next_ys) == 1:
            dec_seq = self.next_ys[0].unsqueeze(1)
        else:
            _, keys = self.sort_scores()
            hyps = [self.get_hypothesis(k) for k in keys]
            hyps = [[2] + h for h in hyps]
            dec_seq = paddle.to_tensor(hyps, dtype='int64')
        return dec_seq

    def get_hypothesis(self, k):
        """ Walk back to construct the full hypothesis. """
        hyp = []
        for j in range(len(self.prev_ks) - 1, -1, -1):
            hyp.append(self.next_ys[j + 1][k])
            k = self.prev_ks[j][k]
        return list(map(lambda x: x.item(), hyp[::-1]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2019/12/04 11:16


""" 区间类

关于区间类，可以参考： https://github.com/AlexandreDecan/python-intervals
但其跟我的业务场景有区别，不太适用，所以这里还是开发了自己的功能库

文档： https://histudy.yuque.com/docs/share/365f3a75-28d0-4595-bc80-5e9d6ab36f71#
"""

import collections
import itertools
import math
import re


class Interval:
    """
    这个类要考虑跟正则的 Match 对象兼容，所以有些细节比较“诡异”
        主要是区间的记录，是用了特殊的 regs 格式
        即虽然说是一个区间，但这个区间可以标记许多子区间
        对应正则里的group(0)，和众多具体的子区间group(1)、group(2)等
        正则里是用regs存储这些区间，所以这里的Interval跟正则Match的regs概念相同
    这里的形式统一为左闭右开
    """
    __slots__ = ('regs',)

    def __init__(self, arg1=None, arg2=None):
        if isinstance(arg1, int) and isinstance(arg2, int):
            # 正常的构建方式
            self.regs = ((arg1, arg2),)
        elif getattr(arg1, 'regs', None):
            # 有regs成员变量，则直接取用（一般是re的Match对象传过来转区间类的）
            self.regs = arg1.regs
        elif arg2 is None and arg1 and len(arg1) == 2 and isinstance(arg1[0], int):
            self.regs = (tuple(arg1),)
        elif arg1:
            # 直接传入区间集
            self.regs = tuple(arg1)
        else:
            # 空区间
            self.regs = None
        self.update()

    def update(self):
        """ 将空区间统一标记为None

        >>> Interval(5, 5).regs  # 返回 None
        >>> Interval(6, 5).regs
        """
        if self.regs and self.regs[0][0] >= self.regs[0][1]:
            self.regs = None

    def start(self, idx=0):
        return self.regs[idx][0] if self.regs else math.inf

    def end(self, idx=0):
        return self.regs[idx][1] if self.regs else -math.inf

    def __bool__(self):
        """
        >>> bool(Interval(5, 6))
        True
        >>> bool(Interval())
        False
        >>> bool(Interval(5, 5))  # 由于标记是左闭右开
        False
        >>> bool(Interval(6, 5))
        False
        """
        if self.regs:  # self.regs存在还不够，区间必须要有长度
            return self.end() > self.start()
        else:
            return False

    def __repr__(self):
        """ 所有区间都是左闭右开！ 第一组(x~y)是主区间，后面跟的是子区间

        >>> Interval()  # 空区间
        []
        >>> Interval(6, 5)  # 空区间
        []
        >>> Interval(4, 8)  # 会将底层的左闭右开区间值，改成左闭右闭的区间值显示，更直观
        [4~7]
        >>> Interval(5, 6)  # 只有单个值的，就不写5~5了，而是简洁性写成5
        [5]
        >>> Interval(((4, 8), (4, 6), (7, 8)))  # 如果有子区间，会在主区间冒号后显示
        [4~7: 4~5 7]
        """
        li = []
        if self.regs:
            li = [(f'{a}~{b - 1}' if b - a > 1 else str(a)) for a, b in self.regs]
            if len(li) > 1: li[0] += ':'
        return '[' + ' '.join(li) + ']'

    def __eq__(self, other):
        """只要主区间标记范围一致就是相等的
        >>> Interval(5, 7) == Interval([(5, 7), (4, 6)])
        True
        >>> Interval(5, 5) == Interval()
        True
        >>> Interval(5, 6) == Interval(5, 7)
        False
        """
        a = self.regs and self.regs[0]
        b = other.regs and other.regs[0]
        return a == b

    def __lt__(self, other):
        """两个区间比大小，只考虑regs[0]，先比第1个值，如果相同再比第2个值

        >>> Interval(4, 6) < Interval(7, 8)
        True
        >>> Interval(4, 5) < Interval(4, 6)
        True
        >>> Interval(4, 6) < Interval(5, 6)
        True
        >>> Interval(4, 6) > Interval(7, 8)  # 虽然只写了<，但python对>也能智能对称操作
        False
        """
        return (self.end() < other.end()) if self.start() == other.start() else (self.start() < other.start())

    def __and__(self, other):
        """
        >>> Interval(4, 7) & Interval(5, 8)
        [5~6]
        >>> Interval(4, 6) & Interval(5, 8)
        [5]
        >>> Interval(4, 6) & Interval(6, 8)
        []

        # 可以和区间集对象运算，会把区间集当成一系列的子区间，将其上下限范围作为主区间来分析
        #    注意和实际的线段点集求并的结果相区别：Intervals([(2, 4), [9, 11]]) & Interval(0, 10)
        >>> Interval(0, 10) & Intervals([(2, 4), [9, 11]])
        [2~9]
        """
        # 如果左右值不合理，类初始化里自带的update会自动变为None
        return Interval(max(self.start(), other.start()), min(self.end(), other.end()))

    def __contains__(self, other):
        """
        :param other: 另一个Interval对象
        :return: self.regs[0] 是否包含了 other.regs[0]

        >>> Interval(6, 8) in Interval(4, 10)
        True
        >>> Interval(2, 7) in Interval(4, 10)
        False
        >>> Intervals([(1,3), (2,4)]) in Interval(1, 4)  # 可以和Intervals混合使用
        True
        >>> Interval(2, 7) not in Interval(4, 10)
        True
        """
        return self.start() <= other.start() and self.end() >= other.end()

    def __or__(self, other):
        """两个regs[0]共有的部分
        :param other: 另一个Interval对象 或 Intervals 区间集对象
        :return: 返回other对象的类型（注意，会丢失所有子区间！），
            如果不存在则regs的值为None
        >>> m1, m2, m3 = Interval(4, 6), Interval(5, 8), Interval(10, 15)
        >>> m1 | m2  # 注意两个子区间会按顺序排
        [4~7: 4~5 5~7]
        >>> m1 | m3
        [4~14: 4~5 10~14]
        >>> m1 | Intervals([(2, 4), (7, 9)])
        [2~8: 2~8 4~5]
        """
        left, right = min(self.start(), other.start()), max(self.end(), other.end())
        a, b = sorted([self.regs[0], (other.start(), other.end())])  # 小的排左边
        return Interval(((left, right), a, b))

    def __add__(self, other):
        """
        >>> Interval(10, 15) + 6
        [16~20]
        """
        if isinstance(other, int):
            regs = [(t[0] + other, t[1] + other) for t in self.regs]
            return Interval(regs)
        else:
            return self | other

    def __sub__(self, other):
        """
        :param other: 另一个Interval对象 或者 Intervals对象
        :return: self.regs[0] 减去 other.regs[0] 后剩余的区间（会丢失子空间集）

        区间减区间：,
        >>> Interval(4, 6) - Interval(5, 8)
        [4]

        # 这种特殊情况会返回含有两个子区间的Interval对象
        >>> Interval(0, 10) - Interval(4, 6)
        [0~9: 0~3 6~9]

        # 这里后者实际区间值并不包含前者，
        #   但实际是按后者的start、end界定的范围作为一个Interval来减的
        >>> Interval(4, 7) - Intervals([(2, 3), (7, 8)])
        []
        """
        if isinstance(other, Intervals):
            other = Interval(other.start(), other.end())
        a, a1, a2 = self, self.start(), self.end()
        b, b1, b2 = other, other.start(), other.end()
        if a1 >= b2 or a2 <= b1:  # a 与 b 不相交
            # 这里不能直接返回a，如果a有子区间，会混淆return值类型
            return Interval(a1, a2)
        else:
            c1, c2 = Interval(a1, b1), Interval(b2, a2)
            if c1 and not c2:
                return c1
            elif c2 and not c1:
                return c2
            elif not c1 and not c2:
                return Interval()
            else:
                return Interval(((a1, a2), (c1.start(), c1.end()), (c2.start(), c2.end())))


class ReMatch(Interval):
    """
    1、伪re._sre.SRE_Match类
        真Match类在re.py的第223行
        有什么办法嘞，标准Match不支持修改成员变量，不支持自定义spes
    2、这个类同时还可以作为“区间”类使用
        有配套的Intervals区间集类，有很刁的各种区间运算功能
    """
    __slots__ = ('regs', 'string', 'pos', 'endpos', 'lastindex', 'lastgroup', 're')

    def __init__(self, regs=None, string=None, pos=0, endpos=None, lastindex=None, lastgroup=None, re=None):
        """Create a new match object.

        :param regs: 区间值
        :param string: 原始的完整字符串内容
        :param pos: 匹配范围开始的位置，一般就是0
        :param endpos: 匹配范围的结束位置，一般就是原字符串长度
        :param lastindex: int，表示有多少个子分组
        :param lastgroup: NoneType，None，The name of the last matched capturing group,
            or None if the group didn’t have a name, or if no group was matched at all.
        :param re: 使用的原正则匹配模式
        """
        if getattr(regs, 'regs', None):
            # 从一个类match对象来初始化
            m = regs
            self.pos = getattr(m, 'pos', None)
            self.endpos = getattr(m, 'endpos', None)
            self.lastindex = getattr(m, 'lastindex', None)
            self.lastgroup = getattr(m, 'lastgroup', None)
            self.re = getattr(m, 're', None)
            self.string = getattr(m, 'string', None)
            self.regs = getattr(m, 'regs', None)
        else:
            self.regs = regs
            self.string = string
            self.pos = pos
            self.endpos = endpos
            self.lastindex = lastindex
            if not self.lastindex and len(self.regs) > 1: self.lastindex = len(self.regs) - 1
            self.lastgroup = lastgroup
            self.re = re
        self.update()

    def group(self, idx=0):
        return self.string[self.regs[idx][0]:self.regs[idx][1]]

    def expand(self, template):
        """Return the string obtained by doing backslash substitution on the
        template string template.

        好像是个输入'\1'可以返回匹配的第1组类似这样的功能

        :type template: T
        :rtype: T
        """
        raise NotImplementedError

    def groups(self, default=None):
        """Return a tuple containing all the subgroups of the match, from 1 up
        to however many groups are in the pattern.

        :rtype: tuple
        """
        return tuple(map(lambda x: self.string[x[0]:x[1]], self.regs[1:]))

    def groupdict(self, default=None):
        """Return a dictionary containing all the named subgroups of the match,
        keyed by the subgroup name.

        :rtype: dict[bytes | unicode, T]
        """
        raise NotImplementedError

    def span(self, group=0):
        """Return a 2-tuple (start, end) for the substring matched by group.

        :type group: int | bytes | unicode
        :rtype: (int, int)
        """
        return self.regs[group][0], self.regs[group][1]


class Intervals:

    def __init__(self, li=None):
        """
        :param li: 若干interval对象
        """
        # 1 matches支持list等类型初始化
        if hasattr(li, 'intervals'):
            li = li.intervals
        if isinstance(li, Intervals):
            self.__dict__ = li.__dict__
        else:
            self.li = []
            if li is None: li = []
            for m in li:
                if not isinstance(m, Interval):
                    m = Interval(m)
                if m: self.li.append(m)  # 只加入非空区间
            self.li.sort()  # 按顺序排
            # 2 生成成员变量
            # self._start = min([m.start() for m in self.li], default=math.inf)
            self._start = min([m.start() for m in self.li[:1]], default=math.inf)
            self._end = max([m.end() for m in self.li], default=-math.inf)

    def start(self):
        """和Interval操作方法尽量对称，头尾也用函数来取，不要用成员变量取"""
        return self._start

    def end(self):
        return self._end

    def _merge_intersect_interval(self, adjacent=True):
        if not self: return []
        li = [self[0]]
        for m in self[1:]:
            if li[-1].end() > m.start() or (adjacent and li[-1].end() == m.start()):
                li[-1] = m | li[-1]  # 如果跟上一个相交，则合并过去
            else:
                li.append(m)  # 否则新建一个区间
        return li

    def merge_intersect_interval(self, adjacent=False):
        """将存在相交的区域进行合并

        :param adjacent: 如果相邻紧接，也进行拼接

        # 注意(1,3)和(2,4)合并为一个区间了
        >>> Intervals([(1, 3), (2, 4), (5, 6)]).merge_intersect_interval(True)
        {[1~3: 1~2 2~3], [5]}
        >>> Intervals([(1, 2), (2, 3)]).merge_intersect_interval(True)
        {[1~2: 1 2]}
        >>> Intervals([(1, 2), (2, 3)]).merge_intersect_interval(adjacent=False)
        {[1], [2]}
        """
        # 因为可能经常被调用，所以要变成static存储
        if adjacent:
            self._li1 = getattr(self, '_li1', None)
            if self._li1 is None:
                self._li1 = Intervals(self._merge_intersect_interval(adjacent))
            return self._li1
        else:
            self._li2 = getattr(self, '_li2', None)
            if self._li2 is None:
                self._li2 = Intervals(self._merge_intersect_interval(adjacent))
            return self._li2

    def true_intersect_subinterval(self, other):
        """判断改区间集，与other区间集，是否存在相交的子区间（真相交，不含子集关系）

        如果一个区间只是self或other独有，或者一个子区间被另一方的一个子区间完全包含，是存在很多优良性质，方便做很多自动化的
        否则，如果存在相交的子区间，就麻烦了

        其实就是想把两个区间完全相同的子区间去掉，然后求交就好了~~

        >>> a, b = Intervals([(1,3), (7,9), (10,12)]), Intervals([(2,4), (7,9), (10, 15)])
        >>> a.true_intersect_subinterval(b)  # 在集合交的基础上，去掉完全包含的情况
        {[2]}
        """
        # 0 区间 转 区间集
        if isinstance(other, Interval):
            other = Intervals([other])

        # 1 区间集和区间集做相交运算，生成一个新的区间集
        """假设a、b都是从左到右按顺序取得，所以可以对b的遍历进行一定过滤简化"""
        A = self.merge_intersect_interval()
        B = other.merge_intersect_interval()
        li, k = [], 0
        for a in A:
            for j in range(k, len(B)):
                b = B[j]
                x1, y1, x2, y2 = a.start(), a.end(), b.start(), b.end()
                if y2 <= x1:
                    # B[0~j]都在a前面，在后续的a中，可以直接从B[j]开始找
                    k = j
                elif x2 >= y1:  # b已经到a右边，后面的b不用再找了，不会有相交
                    break
                elif (x2 < x1 < y2 < y1) or (x1 < x2 < y1 < y2):  # 严格相交，非子集关系
                    li.append(a & b)
        return Intervals(li)

    def sub(self, s, repl, *, out_repl=None, adjacent=False) -> str:
        r"""
        :param repl: 替换的规则函数
            暂不支持和正则等价的字符串替换规则表达
                这个得找技巧，用re现成的功能代码，不可能自己暴力解析
        :param out_repl: 对范围外若有处理需要，可以自定义处理函数
        :param s: 要处理的文本串
        原版 re.sub 还有 count 和 flags 参数，这里难开发，暂时先不做这个接口
        :return:

        >>> s = '0123456789'
        >>> inters = Intervals([(2, 5), (7, 8)])
        >>> inters.sub(s, lambda m: 'b')
        '01b56b89'
        >>> inters.sub(s, lambda m: 'b', out_repl=lambda m: 'a')
        'ababa'
        >>> inters.sub(s, 'b')
        '01b56b89'
        >>> inters.sub(s, 'b', out_repl='a')
        'ababa'
        >>> inters.sub(s, lambda m: ' ' + ''.join(reversed(m.group())) + ' ')
        '01 432 56 7 89'
        >>> inters.sub(s, lambda m: ' ' + ''.join(reversed(m.group())) + ' ', out_repl=lambda m: str(len(m.group())))
        '2 432 2 7 2'
        """
        res, idx = [], 0

        def str2func(a):
            # TODO，如果是str类型，应该要处理字符串标记中的编组和转义等信息的
            return (lambda s: a) if isinstance(a, str) else a

        repl, out_repl = str2func(repl), str2func(out_repl)

        def func1(regs):
            return repl(ReMatch(regs, s, 0, len(s)))  # 构造伪match类并传入

        def func2(start_, end_):
            if out_repl:
                return out_repl(ReMatch(((start_, end_),), s, 0, len(s)))
            else:
                return s[start_:end_]

        for inter in self.merge_intersect_interval(adjacent=adjacent):
            # 匹配范围外的文本处理
            if inter.start() > idx:
                res.append(func2(idx, inter.start()))
            # 匹配范围内的处理
            res.append(func1(inter.regs))
            idx = inter.end()
        if idx < len(s): res.append(func2(idx, len(s)))
        return ''.join(res)

    def replace(self, s, arg1, arg2=None, *, out_repl=lambda s: s, adjacent=False) -> str:
        r"""类似sub函数，但是对两个自定义函数传入的是普通字符串类型，而不是match对象

        :param arg1: 可以输入一个自定义函数
        :param arg2: 可以配合arg1使用，功能同str.replace(arg1, arg2)
        :param adjacent: 替换的时候，为了避免混乱出错，是先要合并重叠的区间集的
            这里有个adjacent参数，True表示邻接的区间会合并，反之则不会合并临接区间

        >>> s = '0123456789'
        >>> inters = Intervals([(2, 5), (7, 8)])
        >>> inters.replace(s, lambda s: 'b')
        '01b56b89'
        >>> inters.replace(s, lambda s: 'b', out_repl=lambda s: 'a')
        'ababa'
        >>> inters.replace(s, 'b')
        '01b56b89'
        >>> inters.replace(s, '2', 'b')
        '01b3456789'
        >>> inters.replace(s, lambda s: ' ' + ''.join(reversed(s)) + ' ', out_repl=lambda s: str(len(s)))
        '2 432 2 7 2'
        """
        res, idx = [], 0

        def str2func(a):
            return (lambda s: a) if isinstance(a, str) else a

        repl, out_repl = str2func(arg1), str2func(out_repl)
        if arg2:
            repl = lambda a: a.replace(arg1, arg2)

        for inter in self.merge_intersect_interval(adjacent=adjacent):
            # 匹配范围外的文本处理
            if inter.start() >= idx:
                res.append(out_repl(s[idx:inter.start()]))
                idx = inter.end()
            # 匹配范围内的处理
            res.append(repl(s[inter.start():inter.end()]))
        if idx < len(s): res.append(out_repl(s[idx:]))
        return ''.join(res)

    def __bool__(self):
        """
        >>> bool(Intervals())
        False
        >>> bool(Intervals([(1, 2), (4, 5)]))
        True
        >>> bool(Intervals([(2, 1), (5, 4)]))
        False
        """
        return bool(self.li)

    def __getitem__(self, item):
        return self.li[item]

    def __iter__(self):
        for m in self.li:
            yield m

    def __len__(self):
        return len(self.li)

    def __repr__(self):
        return '{' + ', '.join([str(m) for m in self.li]) + '}'

    def __eq__(self, other):
        """"数量相等，且每个Interval也相等

        即只考虑强相等，不考虑“弱相等”。
        例如两个区间集虽然数量不同，但使用merge_intersect_interval后，再比较可能就是一样的。

        >>> Intervals([(1,2), (3,5)]) == Intervals([(1,2), (3,5)])
        True
        >>> Intervals([(1,2), (3,5)]) == Intervals([(1,2), (3,4), (4,5)])
        False
        >>> Intervals([(1,2), (3,5)]) == Intervals([(1,2), (3,4), (4,5)]).merge_intersect_interval(True)
        True
        """
        if len(self) != len(other): return False
        for i in range(len(self)):
            if self[i] != other[i]:
                return False
        return True

    def __invert__(self, maxn=None):
        """取反区间集的补集
        注意这样会丢失所有区间的子区间标记

        >>> ~Intervals([(1, 3), (4, 6), (8, 10)])  # 区间取反操作
        {[0], [3], [6~7]}
        >>> ~Intervals([])  # 区间取反操作
        {}
        """
        # 1 要先把有相交的区间合并了
        itvs = self.merge_intersect_interval()

        # 2 辅助变量
        li = []
        if maxn is None: maxn = itvs.end()  # 计算出坐标上限

        # 3 第1个区间是否从0开始
        if len(itvs) and itvs[0].start() == 0:
            idx = itvs[0].end()
            i = 1
        else:
            i = idx = 0

        # 4 循环取得新的区间值
        for m in itvs[i:]:
            li.append(Interval(idx, m.start()))
            idx = m.end()

        # 5 最后一个区间特判
        if idx != maxn: li.append(Interval(idx, maxn))
        res = Intervals(li)
        return res

    def invert(self, maxn=None):
        """
        >>> Intervals([(1, 3), (4, 6), (8, 10)]).invert(20)
        {[0], [3], [6~7], [10~19]}
        """
        return self.__invert__(maxn)

    def __and__(self, other):
        r"""
        # 区间集和单个区间的相交运算：
        >>> Intervals([(2, 4), (9, 11)]) & Interval(0, 10)
        {[2~3], [9]}

        # 区间集和区间集的相交运算：
        >>> Intervals([(1, 5), (6, 8)]) & Intervals([(2, 7), (7, 9)])
        {[2~4], [6], [7]}

        >>> Intervals([(2, 11)]) & Intervals()
        {}
        >>> Intervals() & Intervals([(2, 11)])
        {}
        """
        # 0 区间 转 区间集
        if isinstance(other, Interval):
            other = Intervals([other])

        # 1 区间集和区间集做相交运算，生成一个新的区间集
        """假设a、b都是从左到右按顺序取得，所以可以对b的遍历进行一定过滤简化"""
        A = self.merge_intersect_interval()
        B = other.merge_intersect_interval()
        li, k = [], 0
        for a in A:
            for j in range(k, len(B)):
                b = B[j]
                if b.end() <= a.start():
                    # B[0~j]都在a前面，在后续的a中，可以直接从B[j]开始找
                    k = j
                elif b.start() >= a.end():  # b已经到a右边，后面的b不用再找了，不会有相交
                    break
                else:  # 可能有相交
                    li.append(a & b)
        return Intervals(li)

    def __contains__(self, other):
        r"""
        >>> Interval(3, 5) in Intervals([(2, 6)])
        True
        >>> Interval(3, 5) in Intervals([(0, 4)])
        False
        >>> Intervals([(1, 2), (3, 4)]) in Intervals([(0, 3), (3, 5)])
        True
        >>> Interval(3, 5) not in Intervals([(2, 6)])
        False

        这里具体实现，可以双循环暴力，但考虑区间集顺序性，其实只要双指针同时往前找就好了，
            设几个条件去对循环进行优化，跳出，能大大提高效率
        """
        # 1 区间集 是否包含 区间，转为 区间集 是否包含 区间集 处理
        if isinstance(other, (Interval, list, tuple)):
            other = Intervals([other])

        # 2 合并相交区域
        A = self.merge_intersect_interval()
        B = other.merge_intersect_interval()

        # 3 看是否每个b，都能在A中找到一个a包含它
        i = 0
        for b in B:
            for j in range(i, len(A)):
                if b in A[j]:
                    # A[j-1]前面都不能包含b，但是A[j]能包含b的，后面的b，A[j-1]也一定包含不到
                    i = j
                    break
                elif A[j].start() > b.end():
                    # 后续的a左边都已经比b的右边更大，就不用再找了，肯定都不会有相交的了
                    return False
            else:  # 找到一个b，在A中不包含它
                return False
        return True

    def __or__(self, other):
        r"""区间集相加运算，合成一个新的区间集对象（会丢失所有子区间）

        出现相交的元素会合成一个新的元素，避免区间集中存在相交的两个元素，在sub时会出bug
        >>> Intervals([(2, 4), (5, 7)]) | Interval(1, 3)
        {[1~3: 1~2 2~3], [5~6]}
        >>> Intervals([(2, 4), (5, 7)]) | Intervals([(1, 3), (6, 9)])
        {[1~3: 1~2 2~3], [5~8: 5~6 6~8]}

        >>> Intervals([(1, 3), (6, 9)]) + 3
        {[4~5], [9~11]}
        """
        if isinstance(other, Interval):
            other = Intervals([other])
        else:
            other = Intervals(other)
        return Intervals(self.li + other.li).merge_intersect_interval()

    def __add__(self, other):
        if isinstance(other, int):
            li = [x + other for x in self.li]
            return Intervals(li)
        else:
            return self | other

    def __sub__(self, other):
        """区间集减法操作（注意跟Interval减法操作是有区别的）
        对于任意的 a ∈ self，更新 a = a - {b | b ∈ other}

        >>> Intervals([(0, 10)]) - Interval(4, 6)
        {[0~3], [6~9]}
        >>> Intervals([(0, 10)]) - Interval(8, 12)
        {[0~7]}

        >>> Intervals([(0, 10), (20, 30)]) - Intervals([(0, 5), (15, 25)])
        {[5~9], [25~29]}
        >>> Intervals([(0, 10), (20, 30)]) - Intervals([(2, 5), (7, 12), (25, 27)])
        {[0~1], [5~6], [20~24], [27~29]}
        """
        # 1 
        if isinstance(other, Interval):
            other = Intervals([other])

        # a - b，一个a可能会拆成a1,a2两段，此时左边的a1可以继续处理，
        #   但是a2要加到堆栈A，留作下一轮处理，所以A要用栈结构处理
        A = list(reversed(self.merge_intersect_interval().li))

        B = other.merge_intersect_interval()

        # 2 
        li, k = [], 0
        while A:
            a = A.pop()
            for j in range(k, len(B)):
                b = B[j]
                if b.end() < a.start():
                    k = j
                elif a.end() < a.start():
                    break
                else:
                    c = a - b
                    if not c:  # a已经被减光，就直接跳出循环了
                        a = Interval()
                        break
                    elif len(c.regs) == 1:
                        a = c
                    else:  # 如果 a - c 变成了两段，则左边a1继续处理，右边a2加入A下轮处理
                        a = Interval(c.regs[1])
                        A.append(Interval(c.regs[2]))
            if a: li.append(a)

        return Intervals(li).merge_intersect_interval()


def iter_intervals(arg):
    """从多种类区间类型来构造Interval对象，返回值可能有多组"""

    def judge_range(t):
        return hasattr(t, '__len__') and len(t) == 2 and isinstance(t[0], int) and isinstance(t[1], int)

    if hasattr(arg, 'regs'):
        yield Interval(arg)
    elif judge_range(arg):
        yield Interval(arg)
    elif isinstance(arg, Interval):
        yield arg
    elif isinstance(arg, Intervals):
        for i in range(len(arg)):
            yield arg[i]
    elif hasattr(arg, '__len__') and len(arg) and judge_range(arg[0]):
        for i in range(len(arg)):
            yield Interval(arg[i])
    elif isinstance(arg, collections.Iterable):
        for t in list(arg):
            yield t


def highlight_intervals(content, intervals, colors=None, background=True, showmath=False):
    """文本匹配可视化
    获得高亮显示的匹配区间的html代码

    :param content：需要展示的文本内容
    :param intervals: 输入一个数组，数组的每个元素支持单区间或区间集相关类
        Interval、re正则的Match对象、(4, 10)
        Intervals、[(2,4), (6,8)]

        请自行保证区间嵌套语法正确性，本函数不检查处理嵌套混乱错误问题
    :param colors: 一个数组，和intervals一一对应，轮询使用的颜色
        默认值为： ['red']
    :param background:
        True，使用背景色
        False，不使用背景色，而是字体颜色
    :param showmath:
        True，渲染公式
        False，不渲染公式，只以文本展示
    """
    # 1 存储要插入的html样式
    from collections import defaultdict
    import html

    d = defaultdict(str)

    # 2 其他所有子组从颜色列表取颜色清单，每组一个颜色
    if colors is None:
        colors = ('red',)
    elif isinstance(colors, str):
        colors = (colors,)
    n = len(colors)
    for i, arg in enumerate(intervals):
        color = colors[i % n]
        for interval in iter_intervals(arg):
            l, r = interval.start(), interval.end()
            if background:
                d[l] = d[l] + f'<span style="background-color: {color}">'
                d[r] = '</span>' + d[r]
            else:
                d[l] = d[l] + f'<font color={color}>'
                d[r] = '</font>' + d[r]

    # 3 拼接最终的html代码
    # （1）配一个特殊的pre样式
    head = r"""<!DOCTYPE html>
<html>

<head>
<script src="https://cdn.jsdelivr.net/gh/google/code-prettify@master/loader/run_prettify.js"></script>
"""
    if showmath:
        head += """<script src="https://a.cdn.histudy.com/lib/config/mathjax_config-klxx.js?v=1.1"></script>
<script type="text/javascript" async="" src="https://a.cdn.histudy.com/lib/mathjax/2.7.1/MathJax/MathJax.js?config=TeX-AMS-MML_SVG">
MathJax.Hub.Config(MATHJAX_KLXX_CONFIG);
</script>
        """
    head += r"""</head>

<body>
<pre class="prettyprint nocode linenums">
"""
    tail = '</pre>\n</body>\n</html>'

    # （2）辅助变量
    res = [head]
    s = content
    idxs = sorted(d.keys())  # 按顺序取需要插入的下标

    # （3）拼接
    if idxs: res.append(s[:idxs[0]])
    for i in range(1, len(idxs)):
        res.append(d[idxs[i - 1]])
        res.append(html.escape(s[idxs[i - 1]:idxs[i]]))
    if idxs:  # 最后一个标记
        res.append(d[idxs[-1]])
        res.append(s[idxs[-1]:])
    res.append(tail)

    return ''.join(res)


class StrIdxBack:
    r"""字符串删除部分干扰字符后，对新字符串匹配并回溯找原字符串的下标

    >>> ob = StrIdxBack('bxx  ax xbxax')
    >>> ob.delchars(r'[ x]+')
    >>> ob  # 删除空格、删除字符x
    baba
    >>> print(ob.idx)  # keystr中与原字符串对应位置：(0, 5, 9, 11)
    (0, 5, 9, 11)
    >>> m = re.match(r'b(ab)', ob.keystr)
    >>> m = ob.matchback(m)
    >>> m.group(1)
    'ax xb'
    >>> ob.search('ab')  # 找出原字符串中内容：'ax xb'
    'ax xb'
    """

    def __init__(self, s):
        self.oristr = s
        self.idx = tuple(range(len(s)))  # 存储还保留着内容的下标
        self.keystr = s

    def delchars(self, pattern, flags=0):
        """模仿正则的替换语法
        但是不用输入替换目标s，以及目标格式，因为都是删除操作

        利用正则可以知道被删除的是哪个区间范围
        >>> ob = StrIdxBack('abc123df4a'); ob.delchars(r'\d+'); str(ob)
        'abcdfa'
        >>> ob.idx
        (0, 1, 2, 6, 7, 9)
        """
        k = 0
        idxs = []

        def repl(m):
            nonlocal k, idxs
            idxs.append(self.idx[k:m.start(0)])
            k = m.end(0)
            return ''

        self.keystr = re.sub(pattern, repl, self.keystr, flags=flags)
        idxs.append(self.idx[k:])
        self.idx = tuple(itertools.chain(*idxs))

    def compare_newstr(self, limit=300):
        r"""比较直观的比较字符串前后变化

        newstr相对于oldnew作展开，比较直观的显示字符串前后变化差异
        >>> ob = StrIdxBack('abab'); ob.delchars('b'); ob.compare_newstr()
        'a a '
        """
        s1 = self.oristr
        dd = set(self.idx)

        s2 = []
        k = 0
        for i in range(min(len(s1), limit)):
            if i in dd:
                s2.append(s1[i])
                k += 1
            else:
                if ord(s1[i]) < 128:
                    if s1[i] == ' ':  # 原来是空格的，删除后要用_表示
                        s2.append('_')
                    else:  # 原始不是空格的，可以用空格表示已被删除
                        s2.append(' ')
                else:  # 中文字符要用两个空格表示才能对齐
                    s2.append('  ')
        s2 = ''.join(s2)
        s2 = s2.replace('\n', r'\n')

        return s2

    def compare(self, limit=300):
        """比较直观的比较字符串前后变化"""
        s1 = self.oristr

        s1 = s1.replace('\n', r'\n')[:limit]
        s2 = self.compare_newstr(limit)

        return s1 + '\n' + s2 + '\n'

    def matchback(self, m):
        """输入一个keystr匹配的match对象，将其映射回oristr的match对象"""
        regs = []
        for rs in getattr(m, 'regs'):
            regs.append((self.idx[rs[0]], self.idx[rs[1] - 1] + 1))  # 注意右边界的处理有细节
        return ReMatch(regs, self.oristr, m.pos, len(self.oristr), m.lastindex, m.lastgroup, m.re)

    def search(self, pattern):
        """在新字符串上查找模式，但是返回的是原字符串的相关下标数据"""
        m = re.search(pattern, self.keystr)
        if m:
            m = self.matchback(m)  # pycharm这里会提示m没有regs的成员变量，其实是正常的，没问题
            return m.group()
        else:
            return ''

    def __repr__(self):
        """返回处理后当前的新字符串"""
        return self.keystr



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/clovaai/deep-text-recognition-benchmark/blob/master/modules/transformation.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import paddle
from paddle import nn, ParamAttr
from paddle.nn import functional as F
import numpy as np


class ConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 groups=1,
                 act=None,
                 name=None):
        super(ConvBNLayer, self).__init__()
        self.conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2,
            groups=groups,
            weight_attr=ParamAttr(name=name + "_weights"),
            bias_attr=False)
        bn_name = "bn_" + name
        self.bn = nn.BatchNorm(
            out_channels,
            act=act,
            param_attr=ParamAttr(name=bn_name + '_scale'),
            bias_attr=ParamAttr(bn_name + '_offset'),
            moving_mean_name=bn_name + '_mean',
            moving_variance_name=bn_name + '_variance')

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x


class LocalizationNetwork(nn.Layer):
    def __init__(self, in_channels, num_fiducial, loc_lr, model_name):
        super(LocalizationNetwork, self).__init__()
        self.F = num_fiducial
        F = num_fiducial
        if model_name == "large":
            num_filters_list = [64, 128, 256, 512]
            fc_dim = 256
        else:
            num_filters_list = [16, 32, 64, 128]
            fc_dim = 64

        self.block_list = []
        for fno in range(0, len(num_filters_list)):
            num_filters = num_filters_list[fno]
            name = "loc_conv%d" % fno
            conv = self.add_sublayer(
                name,
                ConvBNLayer(
                    in_channels=in_channels,
                    out_channels=num_filters,
                    kernel_size=3,
                    act='relu',
                    name=name))
            self.block_list.append(conv)
            if fno == len(num_filters_list) - 1:
                pool = nn.AdaptiveAvgPool2D(1)
            else:
                pool = nn.MaxPool2D(kernel_size=2, stride=2, padding=0)
            in_channels = num_filters
            self.block_list.append(pool)
        name = "loc_fc1"
        stdv = 1.0 / math.sqrt(num_filters_list[-1] * 1.0)
        self.fc1 = nn.Linear(
            in_channels,
            fc_dim,
            weight_attr=ParamAttr(
                learning_rate=loc_lr,
                name=name + "_w",
                initializer=nn.initializer.Uniform(-stdv, stdv)),
            bias_attr=ParamAttr(name=name + '.b_0'),
            name=name)

        # Init fc2 in LocalizationNetwork
        initial_bias = self.get_initial_fiducials()
        initial_bias = initial_bias.reshape(-1)
        name = "loc_fc2"
        param_attr = ParamAttr(
            learning_rate=loc_lr,
            initializer=nn.initializer.Assign(np.zeros([fc_dim, F * 2])),
            name=name + "_w")
        bias_attr = ParamAttr(
            learning_rate=loc_lr,
            initializer=nn.initializer.Assign(initial_bias),
            name=name + "_b")
        self.fc2 = nn.Linear(
            fc_dim,
            F * 2,
            weight_attr=param_attr,
            bias_attr=bias_attr,
            name=name)
        self.out_channels = F * 2

    def forward(self, x):
        """
           Estimating parameters of geometric transformation
           Args:
               image: input
           Return:
               batch_C_prime: the matrix of the geometric transformation
        """
        B = x.shape[0]
        i = 0
        for block in self.block_list:
            x = block(x)
        x = x.squeeze(axis=2).squeeze(axis=2)
        x = self.fc1(x)

        x = F.relu(x)
        x = self.fc2(x)
        x = x.reshape(shape=[-1, self.F, 2])
        return x

    def get_initial_fiducials(self):
        """ see RARE paper Fig. 6 (a) """
        F = self.F
        ctrl_pts_x = np.linspace(-1.0, 1.0, int(F / 2))
        ctrl_pts_y_top = np.linspace(0.0, -1.0, num=int(F / 2))
        ctrl_pts_y_bottom = np.linspace(1.0, 0.0, num=int(F / 2))
        ctrl_pts_top = np.stack([ctrl_pts_x, ctrl_pts_y_top], axis=1)
        ctrl_pts_bottom = np.stack([ctrl_pts_x, ctrl_pts_y_bottom], axis=1)
        initial_bias = np.concatenate([ctrl_pts_top, ctrl_pts_bottom], axis=0)
        return initial_bias


class GridGenerator(nn.Layer):
    def __init__(self, in_channels, num_fiducial):
        super(GridGenerator, self).__init__()
        self.eps = 1e-6
        self.F = num_fiducial

        name = "ex_fc"
        initializer = nn.initializer.Constant(value=0.0)
        param_attr = ParamAttr(
            learning_rate=0.0, initializer=initializer, name=name + "_w")
        bias_attr = ParamAttr(
            learning_rate=0.0, initializer=initializer, name=name + "_b")
        self.fc = nn.Linear(
            in_channels,
            6,
            weight_attr=param_attr,
            bias_attr=bias_attr,
            name=name)

    def forward(self, batch_C_prime, I_r_size):
        """
        Generate the grid for the grid_sampler.
        Args:
            batch_C_prime: the matrix of the geometric transformation
            I_r_size: the shape of the input image
        Return:
            batch_P_prime: the grid for the grid_sampler
        """
        C = self.build_C_paddle()
        P = self.build_P_paddle(I_r_size)

        inv_delta_C_tensor = self.build_inv_delta_C_paddle(C).astype('float32')
        P_hat_tensor = self.build_P_hat_paddle(
            C, paddle.to_tensor(P)).astype('float32')

        inv_delta_C_tensor.stop_gradient = True
        P_hat_tensor.stop_gradient = True

        batch_C_ex_part_tensor = self.get_expand_tensor(batch_C_prime)

        batch_C_ex_part_tensor.stop_gradient = True

        batch_C_prime_with_zeros = paddle.concat(
            [batch_C_prime, batch_C_ex_part_tensor], axis=1)
        batch_T = paddle.matmul(inv_delta_C_tensor, batch_C_prime_with_zeros)
        batch_P_prime = paddle.matmul(P_hat_tensor, batch_T)
        return batch_P_prime

    def build_C_paddle(self):
        """ Return coordinates of fiducial points in I_r; C """
        F = self.F
        ctrl_pts_x = paddle.linspace(-1.0, 1.0, int(F / 2), dtype='float64')
        ctrl_pts_y_top = -1 * paddle.ones([int(F / 2)], dtype='float64')
        ctrl_pts_y_bottom = paddle.ones([int(F / 2)], dtype='float64')
        ctrl_pts_top = paddle.stack([ctrl_pts_x, ctrl_pts_y_top], axis=1)
        ctrl_pts_bottom = paddle.stack([ctrl_pts_x, ctrl_pts_y_bottom], axis=1)
        C = paddle.concat([ctrl_pts_top, ctrl_pts_bottom], axis=0)
        return C  # F x 2

    def build_P_paddle(self, I_r_size):
        I_r_height, I_r_width = I_r_size
        I_r_grid_x = (paddle.arange(
            -I_r_width, I_r_width, 2, dtype='float64') + 1.0
                      ) / paddle.to_tensor(np.array([I_r_width]))

        I_r_grid_y = (paddle.arange(
            -I_r_height, I_r_height, 2, dtype='float64') + 1.0
                      ) / paddle.to_tensor(np.array([I_r_height]))

        # P: self.I_r_width x self.I_r_height x 2
        P = paddle.stack(paddle.meshgrid(I_r_grid_x, I_r_grid_y), axis=2)
        P = paddle.transpose(P, perm=[1, 0, 2])
        # n (= self.I_r_width x self.I_r_height) x 2
        return P.reshape([-1, 2])

    def build_inv_delta_C_paddle(self, C):
        """ Return inv_delta_C which is needed to calculate T """
        F = self.F
        hat_eye = paddle.eye(F, dtype='float64')  # F x F
        hat_C = paddle.norm(
            C.reshape([1, F, 2]) - C.reshape([F, 1, 2]), axis=2) + hat_eye
        hat_C = (hat_C**2) * paddle.log(hat_C)
        delta_C = paddle.concat(  # F+3 x F+3
            [
                paddle.concat(
                    [paddle.ones(
                        (F, 1), dtype='float64'), C, hat_C], axis=1),  # F x F+3
                paddle.concat(
                    [
                        paddle.zeros(
                            (2, 3), dtype='float64'), paddle.transpose(
                                C, perm=[1, 0])
                    ],
                    axis=1),  # 2 x F+3
                paddle.concat(
                    [
                        paddle.zeros(
                            (1, 3), dtype='float64'), paddle.ones(
                                (1, F), dtype='float64')
                    ],
                    axis=1)  # 1 x F+3
            ],
            axis=0)
        inv_delta_C = paddle.inverse(delta_C)
        return inv_delta_C  # F+3 x F+3

    def build_P_hat_paddle(self, C, P):
        F = self.F
        eps = self.eps
        n = P.shape[0]  # n (= self.I_r_width x self.I_r_height)
        # P_tile: n x 2 -> n x 1 x 2 -> n x F x 2
        P_tile = paddle.tile(paddle.unsqueeze(P, axis=1), (1, F, 1))
        C_tile = paddle.unsqueeze(C, axis=0)  # 1 x F x 2
        P_diff = P_tile - C_tile  # n x F x 2
        # rbf_norm: n x F
        rbf_norm = paddle.norm(P_diff, p=2, axis=2, keepdim=False)

        # rbf: n x F
        rbf = paddle.multiply(
            paddle.square(rbf_norm), paddle.log(rbf_norm + eps))
        P_hat = paddle.concat(
            [paddle.ones(
                (n, 1), dtype='float64'), P, rbf], axis=1)
        return P_hat  # n x F+3

    def get_expand_tensor(self, batch_C_prime):
        B, H, C = batch_C_prime.shape
        batch_C_prime = batch_C_prime.reshape([B, H * C])
        batch_C_ex_part_tensor = self.fc(batch_C_prime)
        batch_C_ex_part_tensor = batch_C_ex_part_tensor.reshape([-1, 3, 2])
        return batch_C_ex_part_tensor


class TPS(nn.Layer):
    def __init__(self, in_channels, num_fiducial, loc_lr, model_name):
        super(TPS, self).__init__()
        self.loc_net = LocalizationNetwork(in_channels, num_fiducial, loc_lr,
                                           model_name)
        self.grid_generator = GridGenerator(self.loc_net.out_channels,
                                            num_fiducial)
        self.out_channels = in_channels

    def forward(self, image):
        image.stop_gradient = False
        batch_C_prime = self.loc_net(image)
        batch_P_prime = self.grid_generator(batch_C_prime, image.shape[2:])
        batch_P_prime = batch_P_prime.reshape(
            [-1, image.shape[2], image.shape[3], 2])
        batch_I_r = F.grid_sample(x=image, grid=batch_P_prime)
        return batch_I_r



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/09/07 10:21


import win32com.client as win32
import pythoncom


def get_win32_app(name, visible=False):
    """ 启动可支持pywin32自动化处理的应用

    Args:
        str name: 应用名称，不区分大小写，比如word, excel, powerpoint, onenote
            不带'.'的情况下，会自动添加'.Application'的后缀
        visible: 应用是否可见

    Returns: app

    """
    # 1 name
    name = name.lower()
    if '.' not in name:
        name += '.application'

    # 2 app
    # 这里可能还有些问题，不同的应用，机制不太一样，后面再细化完善吧
    try:
        app = win32.GetActiveObject(f'{name}')  # 不能关联到普通方式打开的应用。但代码打开的应用都能找得到。
    except pythoncom.com_error:
        app = win32.gencache.EnsureDispatch(f'{name}')
        # 还有种常见的初始化方法，是 win32com.client.Dispatch和win32com.client.dynamic.Dispatch
        # from win32com.client.dynamic import Disypatch

    if visible is not None:
        app.Visible = visible

    return app



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from shapely.geometry import Polygon
"""
:param det_x: [1, N] Xs of detection's vertices 
:param det_y: [1, N] Ys of detection's vertices
:param gt_x: [1, N] Xs of groundtruth's vertices
:param gt_y: [1, N] Ys of groundtruth's vertices

##############
All the calculation of 'AREA' in this script is handled by:
1) First generating a binary mask with the polygon area filled up with 1's
2) Summing up all the 1's
"""


def area(x, y):
    polygon = Polygon(np.stack([x, y], axis=1))
    return float(polygon.area)


def approx_area_of_intersection(det_x, det_y, gt_x, gt_y):
    """
    This helper determine if both polygons are intersecting with each others with an approximation method.
    Area of intersection represented by the minimum bounding rectangular [xmin, ymin, xmax, ymax]
    """
    det_ymax = np.max(det_y)
    det_xmax = np.max(det_x)
    det_ymin = np.min(det_y)
    det_xmin = np.min(det_x)

    gt_ymax = np.max(gt_y)
    gt_xmax = np.max(gt_x)
    gt_ymin = np.min(gt_y)
    gt_xmin = np.min(gt_x)

    all_min_ymax = np.minimum(det_ymax, gt_ymax)
    all_max_ymin = np.maximum(det_ymin, gt_ymin)

    intersect_heights = np.maximum(0.0, (all_min_ymax - all_max_ymin))

    all_min_xmax = np.minimum(det_xmax, gt_xmax)
    all_max_xmin = np.maximum(det_xmin, gt_xmin)
    intersect_widths = np.maximum(0.0, (all_min_xmax - all_max_xmin))

    return intersect_heights * intersect_widths


def area_of_intersection(det_x, det_y, gt_x, gt_y):
    p1 = Polygon(np.stack([det_x, det_y], axis=1)).buffer(0)
    p2 = Polygon(np.stack([gt_x, gt_y], axis=1)).buffer(0)
    return float(p1.intersection(p2).area)


def area_of_union(det_x, det_y, gt_x, gt_y):
    p1 = Polygon(np.stack([det_x, det_y], axis=1)).buffer(0)
    p2 = Polygon(np.stack([gt_x, gt_y], axis=1)).buffer(0)
    return float(p1.union(p2).area)


def iou(det_x, det_y, gt_x, gt_y):
    return area_of_intersection(det_x, det_y, gt_x, gt_y) / (
        area_of_union(det_x, det_y, gt_x, gt_y) + 1.0)


def iod(det_x, det_y, gt_x, gt_y):
    """
    This helper determine the fraction of intersection area over detection area
    """
    return area_of_intersection(det_x, det_y, gt_x, gt_y) / (
        area(det_x, det_y) + 1.0)



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/whai362/PSENet/blob/python3/models/head/psenet_head.py
"""

from paddle import nn


class PSEHead(nn.Layer):
    def __init__(self, in_channels, hidden_dim=256, out_channels=7, **kwargs):
        super(PSEHead, self).__init__()
        self.conv1 = nn.Conv2D(
            in_channels, hidden_dim, kernel_size=3, stride=1, padding=1)
        self.bn1 = nn.BatchNorm2D(hidden_dim)
        self.relu1 = nn.ReLU()

        self.conv2 = nn.Conv2D(
            hidden_dim, out_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, x, **kwargs):
        out = self.conv1(x)
        out = self.relu1(self.bn1(out))
        out = self.conv2(out)
        return {'maps': out}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/10/08 22:26

import re

from pyxllib.text.pupil import grp_bracket, continuous_zero
from pyxllib.debug.specialist import browser


class TexTabular:
    @classmethod
    def parse_multirow(cls, s, brace_text_only=True):
        r"""

        :param brace_text_only: 只取花括号里面的内容
            如果为False，会把multirow外部的内容做拼接

        multirow 和 multicolumn 的不同是，第1、2个花括号后面可以有可选参数。
        第2个花括号如果内容是*，可以省略。
        两个[]的内容先省略，不做分析处理

        注意：这里会取出前后缀内容！业务需要，防止bug，不过这种概率很小

        >>> TexTabular.parse_multirow(r'\multirow{2}*{特点}')
        (2, None, '*', None, '特点')
        >>> TexTabular.parse_multirow(r'\multirow{2}{*}{特点}')
        (2, None, '*', None, '特点')
        >>> TexTabular.parse_multirow(r'aa\multirow{2}[bla1]{*}[bla2]{特点}bb', brace_text_only=False)
        (2, 'bla1', '*', 'bla2', 'aa特点bb')

        TODO multirow第一个数字是可以负值的，代表向上合并单元格数，
        """
        square = r'(?:\[(.*?)\])?'  # 可选参数
        m = re.search(r'\\multirow' + grp_bracket(3, inner=True) + square +
                      r'(?:{(.*?)}|(\*))' + square + grp_bracket(5, inner=True), s)
        if not m: return None
        n, bigstructs, width1, width2, fixup, text = m.groups()
        width = width1 or width2
        if not brace_text_only: text = s[:m.start()] + text + s[m.end():]
        # if re.match(r'\d+$', text): text = int(text)  # 如果可以，直接识别为数字

        n = int(n)
        if -1 <= n <= 1:
            n = 1
        elif n > 1:
            pass
        else:
            raise ValueError(f'{s} 不支持解析multirow第一个值为负数，向上合并单元格的情况')

        return n, bigstructs, width, fixup, text

    @classmethod
    def parse_multicolumn(cls, s):
        r"""找出s中第一次出现的满足模式的multicolumn，返回3个关键值

        :returns:
            第1个参数是该合并单元格的尺寸，固定格式： (行数, 列数)，只有一行是也会写'1'

        >>> TexTabular.parse_multicolumn(r'\multicolumn{2}{|c|}{aa\multirow{3}*{特点}bb}')
        ((3, 2), '|c|', 'aa特点bb')
        """
        # 1 基本的模式匹配抓取
        m = re.search(r'\\multicolumn' + grp_bracket(3, inner=True) * 2
                      + grp_bracket(5, inner=True), s)  # 最后层多套下，我怕不够用
        if not m: return None

        # 2 取出参数值
        m, col_align, text = m.groups()
        m = int(m)

        # 3 如果有 multirow
        if 'multirow' in text:
            n, bigstructs, width, fixup, text = cls.parse_multirow(text, brace_text_only=False)
        else:
            n = 1
        # if isinstance(text, str) and re.match(r'\d+$', text): text = int(text)  # 如果可以，直接识别为数字
        return (n, m), col_align, text

    @classmethod
    def parse_align(cls, s):
        r"""解析latex表头的列对齐格式

        latex表头的规则很复杂，这里目前只处理一些较常用的功能点

        :param s: 内容文本
        :return: 不考虑竖线和一些高级对齐格式，暂时返回一个str
            长度是表格列数，每个元素是一个字母存储对齐信息（后续可以扩展更细致的对齐格式信息）

        >>> TexTabular.parse_align('{|c|c|c|c|c|c|c|c|c<{}|c|}')
        'cccccccccc'
        >>> TexTabular.parse_align('{|c|w{6em}|w{23mm}|w{47mm}|w{22mm}|}')
        'cwwww'
        >>> TexTabular.parse_align('cc*{8}{l}')
        'ccllllllll'
        >>> TexTabular.parse_align('|c|')
        'c'
        >>> TexTabular.parse_align('|c|*{2}{m{38mm}<{\\centering}|}')
        'cmm'
        """
        # 展开 *{n}{列格式} 模式
        s = re.sub(r'\*(\d+)', r'*{\1}', s)  # 给*数字加上花括号，不然我的匹配会错
        s = re.sub(r'\*{(\d+)}' + grp_bracket(3, inner=True), lambda m: m.group(2) * int(m.group(1)), s)
        # 删除其他干扰字符
        if s[0] == '{' and s[-1] == '}': s = s[1:-1]  # 去掉头尾 { }
        s = re.sub(r'{.*?}', '', s)
        for char in '|<>!':
            s = s.replace(char, '')
        return s

    @classmethod
    def create_cline(cls, merge_count):
        r"""
        :param merge_count: 一个长度等于表格列数的list，第i位的值存储了第i列累计到当前被合并的格子数
            假设一个3*4的表格，第1、3、4列正常，第2列被合并了
            那么遍历到第二行时，merge_count为 [0, 2, 0, 0]
            遍历到第三行时，merge_count 为 [0, 1, 0, 0]
        :return:

        >>> TexTabular.create_cline([0, 1, 0, 0])
        '\\cline{1-1} \\cline{3-4}'
        """

        s = ''.join([('1' if v else '0') for v in merge_count])
        if s.count('1') == 0: return '\\hline'  # 没有间断，直接用hline命令

        spans = continuous_zero(s)  # 注意返回的区间是从0开始编号，左闭右开的
        li = [f'\\cline{{{span[0] + 1}-{span[1]}}}' for span in spans]
        return ' '.join(li)

    @classmethod
    def create_formats(cls, format_count):
        """ 获得latex表头格式 """

        def count(s):
            """列对齐格式统计，返回最终去用的对齐格式"""
            if not s: return 'l'  # 默认左对齐
            l, c, r = s.count('l'), s.count('c'), s.count('r')
            if l >= c and l >= r:
                return 'l'
            elif c >= l and c >= r:
                return 'c'
            else:
                return 'r'

        formats = [count(x) for x in format_count]
        return '{|' + '|'.join(formats) + '|}'


def browser_latex(text='请输入...'):
    from html import escape
    from pyxllib.text.xmllib import get_jinja_template

    # 致谢：感谢奕本在晓波做的工具基础上，做出的这个简洁版的latex渲染器
    content = get_jinja_template('latex_editor.html').render(text=escape(text))
    browser.html(content)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 14:22

import textwrap
from collections import defaultdict, Counter
import math
import re
import sys

from pyxllib.prog.newbie import typename
from pyxllib.text.pupil import listalign, int2myalphaenum


def natural_sort_key(key):
    """
    >>> natural_sort_key('0.0.43') < natural_sort_key('0.0.43.1')
    True

    >>> natural_sort_key('0.0.2') < natural_sort_key('0.0.12')
    True
    """

    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    return [convert(c) for c in re.split('([0-9]+)', str(key))]


def natural_sort(ls, only_use_digits=False):
    """ 自然排序

    :param only_use_digits: 正常会用数字作为分隔，切割每一部分进行比较
        如果只想比较数值部分，可以only_use_digits=True

    >>> natural_sort(['0.1.12', '0.0.10', '0.0.23'])
    ['0.0.10', '0.0.23', '0.1.12']
    """
    if only_use_digits:
        def func(key):
            return [int(c) for c in re.split('([0-9]+)', str(key)) if c.isdigit()]
    else:
        func = natural_sort_key
    return sorted(ls, key=func)


def argsort(seq):
    # http://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python
    return sorted(range(len(seq)), key=seq.__getitem__)


def make_index_function(li, *, start=0, nan=None):
    """ 返回一个函数，输入值，返回对应下标，找不到时返回 not_found

    :param li: 列表数据
    :param start: 起始下标
    :param nan: 找不到对应元素时的返回值
        注意这里找不到默认不是-1，而是li的长度，这样用于排序时，找不到的默认会排在尾巴

    >>> func = make_index_function(['少儿', '小学', '初中', '高中'])
    >>> sorted(['初中', '小学', '高中'], key=func)
    ['小学', '初中', '高中']

    # 不在枚举项目里的，会统一列在最后面
    >>> sorted(['初中', '小学', '高中', '幼儿'], key=func)
    ['小学', '初中', '高中', '幼儿']
    """
    data = {x: i for i, x in enumerate(li, start=start)}
    if nan is None:
        nan = len(li)

    def warpper(x, default=None):
        if default is None:
            default = nan
        return data.get(x, default)

    return warpper


class ValuesStat:
    """ 一串数值的相关统计分析 """

    def __init__(self, values):
        from statistics import stdev, mean
        self.values = values
        self.n = len(values)
        self.sum = sum(values)
        if self.n:
            self.mean = mean(self.values)
            self.std = stdev(self.values)
            self.min, self.max = min(values), max(values)
        else:
            self.mean = self.std = self.min = self.max = float('nan')

    def __len__(self):
        return self.n

    def summary(self, valfmt='g'):
        """ 输出性能分析报告，data是每次运行得到的时间数组

        :param valfmt: 数值显示的格式
            g是比较智能的一种模式
            也可以用 '.3f'表示保留3位小数

            也可以传入长度5的格式清单，表示 [和、均值、标准差、最小值、最大值] 一次展示的格式
        """
        if isinstance(valfmt, str):
            valfmt = [valfmt] * 5

        if self.n > 1:  # 有多轮，则应该输出些参考统计指标
            ls = [f'总和: {self.sum:{valfmt[0]}}', f'均值标准差: {self.mean:{valfmt[1]}}±{self.std:{valfmt[2]}}',
                  f'总数: {self.n}', f'最小值: {self.min:{valfmt[3]}}', f'最大值: {self.max:{valfmt[4]}}']
            return '\t'.join(ls)
        elif self.n == 1:  # 只有一轮，则简单地输出即可
            return f'{self.sum:{valfmt[0]}}'
        else:
            raise ValueError


class Groups:
    def __init__(self, data):
        """ 分组

        :param data: 输入字典结构直接赋值
            或者其他结构，会自动按相同项聚合

        TODO 显示一些数值统计信息，甚至图表
        TODO 转文本表达，方便bc比较
        """
        if not isinstance(data, dict):
            new_data = dict()
            # 否要要转字典类型，自动从1~n编组
            for k, v in enumerate(data, start=1):
                new_data[k] = v
            data = new_data
        self.data = data  # 字典存原数据
        self.ctr = Counter({k: len(x) for k, x in self.data.items()})  # 计数
        self.stat = ValuesStat(self.ctr.values())  # 综合统计数据

    def __repr__(self):
        ls = []
        for i, (k, v) in enumerate(self.data.items(), start=1):
            ls.append(f'{i}, {k}：{v}')
        return '\n'.join(ls)

    @classmethod
    def groupby(cls, ls, key, ykey=None):
        """
        :param ls: 可迭代等数组类型
        :param key: 映射规则，ls中每个元素都会被归到映射的key组上
            Callable[Any, 不可变类型]
            None，未输入时，默认输入的ls已经是分好组的数据
        :param ykey: 是否对分组后存储的内容y，也做一个函数映射
        :return: dict
        """
        data = defaultdict(list)
        for x in ls:
            k = key(x)
            if ykey:
                x = ykey(x)
            data[k].append(x)
        return cls(data)


def intersection_split(a, b):
    """ 输入两个对象a,b，可以是dict或set类型，list等

    会分析出二者共有的元素值关系
    返回值是 ls1, ls2, ls3, ls4，大部分是list类型，但也有可能遵循原始情况是set类型
        ls1：a中，与b共有key的元素值
        ls2：a中，独有key的元素值
        ls3：b中，与a共有key的元素值
        ls4：b中，独有key的元素值
    """
    # 1 获得集合的key关系
    keys1 = set(a)
    keys2 = set(b)
    keys0 = keys1 & keys2  # 两个集合共有的元素

    # TODO 如果是字典，希望能保序

    # 2 组合出ls1、ls2、ls3、ls4

    def split(t, s, ks):
        """原始元素为t，集合化的值为s，共有key是ks"""
        if isinstance(t, (set, list, tuple)):
            return ks, s - ks
        elif isinstance(t, dict):
            ls1 = sorted(map(lambda x: (x, t[x]), ks), key=lambda x: natural_sort_key(x[0]))
            ls2 = sorted(map(lambda x: (x, t[x]), s - ks), key=lambda x: natural_sort_key(x[0]))
            return ls1, ls2
        else:
            # dprint(type(s))  # s不是可以用来进行集合规律分析的类型
            raise ValueError(f'{type(s)}不是可以用来进行集合规律分析的类型')

    ls1, ls2 = split(a, keys1, keys0)
    ls3, ls4 = split(b, keys2, keys0)
    return ls1, ls2, ls3, ls4


def matchpairs(xs, ys, cmp_func, least_score=sys.float_info.epsilon, *,
               key=None, index=False):
    r""" 匹配两组数据

    :param xs: 第一组数据
    :param ys: 第二组数据
    :param cmp_func: 所用的比较函数，值越大表示两个对象相似度越高
    :param least_score: 允许匹配的最低分，默认必须要大于0
    :param key: 是否需要对xs, ys进行映射后再传入 cmp_func 操作
    :param index: 返回的不是原值，而是下标
    :return: 返回结构[(x1, y1, score1), (x2, y2, score2), ...]，注意长度肯定不会超过min(len(xs), len(ys))

    注意：这里的功能①不支持重复匹配，②任何一个x,y都有可能没有匹配到
        如果每个x必须都要有一个匹配，或者支持重复配对，请到隔壁使用 MatchPairs

    TODO 这里很多中间步骤结果都是很有分析价值的，能改成类，然后支持分析中间结果？
    TODO 这样全量两两比较是很耗性能的，可以加个参数草算，不用精确计算的功能？

    >>> xs, ys = [4, 6, 1, 2, 9, 4, 5], [1, 5, 8, 9, 2]
    >>> cmp_func = lambda x,y: 1-abs(x-y)/max(x,y)
    >>> matchpairs(xs, ys, cmp_func)
    [(1, 1, 1.0), (2, 2, 1.0), (9, 9, 1.0), (5, 5, 1.0), (6, 8, 0.75)]
    >>> matchpairs(ys, xs, cmp_func)
    [(1, 1, 1.0), (5, 5, 1.0), (9, 9, 1.0), (2, 2, 1.0), (8, 6, 0.75)]
    >>> matchpairs(xs, ys, cmp_func, 0.9)
    [(1, 1, 1.0), (2, 2, 1.0), (9, 9, 1.0), (5, 5, 1.0)]
    >>> matchpairs(xs, ys, cmp_func, 0.9, index=True)
    [(2, 0, 1.0), (3, 4, 1.0), (4, 3, 1.0), (6, 1, 1.0)]
    """
    # 0 实际计算使用的是 xs_, ys_
    if key:
        xs_ = [key(x) for x in xs]
        ys_ = [key(y) for y in ys]
    else:
        xs_, ys_ = xs, ys

    # 1 计算所有两两相似度
    n, m = len(xs), len(ys)
    all_pairs = []
    for i in range(n):
        for j in range(m):
            score = cmp_func(xs_[i], ys_[j])
            if score >= least_score:
                all_pairs.append([i, j, score])
    # 按分数权重排序，如果分数有很多相似并列，就只能按先来后到排序啦
    all_pairs = sorted(all_pairs, key=lambda v: (-v[2], v[0], v[1]))

    # 2 过滤出最终结果
    pairs = []
    x_used, y_used = set(), set()
    for p in all_pairs:
        i, j, score = p
        if i not in x_used and j not in y_used:
            if index:
                pairs.append((i, j, score))
            else:
                pairs.append((xs[i], ys[j], score))
            x_used.add(i)
            y_used.add(j)

    return pairs


def get_number_width(n):
    """ 判断数值n的长度

    >>> get_number_width(0)
    Traceback (most recent call last):
    AssertionError
    >>> get_number_width(9)
    1
    >>> get_number_width(10)
    2
    >>> get_number_width(97)
    2
    """
    assert n > 0
    return math.ceil(math.log10(n + 1))


class SearchBase:
    """ 一个dfs、bfs模板类 """

    def __init__(self, root):
        """
        Args:
            root: 根节点
        """
        self.root = root

    def get_neighbors(self, node):
        """ 获得邻接节点，必须要用yield实现，方便同时支持dfs、bfs的使用

        对于树结构而言，相当于获取直接子结点

        这里默认是bs4中Tag规则；不同业务需求，可以重定义该函数
        例如对图结构、board类型，可以在self存储图访问状态，在这里实现遍历四周的功能
        """
        try:
            for node in node.children:
                yield node
        except AttributeError:
            pass

    def dfs_nodes(self, node=None, depth=0):
        """ 返回深度优先搜索得到的结点清单

        :param node: 起始结点，默认是root根节点
        :param depth: 当前node深度
        :return: list，[(node1, depth1), (node2, depth2), ...]
        """
        if not node:
            node = self.root

        ls = [(node, depth)]
        for t in self.get_neighbors(node):
            ls += self.dfs_nodes(t, depth + 1)
        return ls

    def bfs_nodes(self, node=None, depth=0):
        if not node:
            node = self.root

        ls = [(node, depth)]
        i = 0

        while i < len(ls):
            x, d = ls[i]
            nodes = self.get_neighbors(x)
            ls += [(t, d + 1) for t in nodes]
            i += 1

        return ls

    def fmt_node(self, node, depth, *, prefix='    ', show_node_type=False):
        """ node格式化显示 """
        s1 = prefix * depth
        s2 = typename(node) + '，' if show_node_type else ''
        s3 = textwrap.shorten(str(node), 200)
        return s1 + s2 + s3

    def fmt_nodes(self, *, nodes=None, select_depth=None, linenum=False,
                  msghead=True, show_node_type=False, prefix='    '):
        """ 结点清单格式化输出

        :param nodes: 默认用dfs获得结点，也可以手动指定结点
        :param prefix: 缩进格式，默认用4个空格
        :param select_depth: 要显示的深度
            单个数字：获得指定层
            Sequences： 两个整数，取出这个闭区间内的层级内容
        :param linenum：节点从1开始编号
            行号后面，默认会跟一个类似Excel列名的字母，表示层级深度
        :param msghead: 第1行输出一些统计信息
        :param show_node_type:

        Requires
            textwrap：用到shorten
            align.listalign：生成列编号时对齐
        """
        # 1 生成结点清单
        ls = nodes if nodes else self.dfs_nodes()
        total_node = len(ls)
        total_depth = max(map(lambda x: x[1], ls))
        head = f'总节点数：1~{total_node}，总深度：0~{total_depth}'

        # 2 过滤与重新整理ls（select_depth）
        logo = True
        cnt = 0
        tree_num = 0
        if isinstance(select_depth, int):

            for i in range(total_node):
                if ls[i][1] == select_depth:
                    ls[i][1] = 0
                    cnt += 1
                    logo = True
                elif ls[i][1] < select_depth and logo:  # 遇到第1个父节点添加一个空行
                    ls[i] = ''
                    tree_num += 1
                    logo = False
                else:  # 删除该节点，不做任何显示
                    ls[i] = None
            head += f'；挑选出的节点数：{cnt}，所选深度：{select_depth}，树数量：{tree_num}'

        elif hasattr(select_depth, '__getitem__'):
            for i in range(total_node):
                if select_depth[0] <= ls[i][1] <= select_depth[1]:
                    ls[i][1] -= select_depth[0]
                    cnt += 1
                    logo = True
                elif ls[i][1] < select_depth[0] and logo:  # 遇到第1个父节点添加一个空行
                    ls[i] = ''
                    tree_num += 1
                    logo = False
                else:  # 删除该节点，不做任何显示
                    ls[i] = None
            head += f'；挑选出的节点数：{cnt}，所选深度：{select_depth[0]}~{select_depth[1]}，树数量：{tree_num}'
        """注意此时ls[i]的状态，有3种类型
            (node, depth)：tuple类型，第0个元素是node对象，第1个元素是该元素所处层级
            None：已删除元素，但为了后续编号方便，没有真正的移出，而是用None作为标记
            ''：已删除元素，但这里涉及父节点的删除，建议此处留一个空行
        """

        # 3 格式处理
        def mystr(item):
            return self.fmt_node(item[0], item[1], prefix=prefix, show_node_type=show_node_type)

        line_num = listalign(range(1, total_node + 1))
        res = []
        for i in range(total_node):
            if ls[i] is not None:
                if isinstance(ls[i], str):  # 已经指定该行要显示什么
                    res.append(ls[i])
                else:
                    if linenum:  # 增加了一个能显示层级的int2excel_col_name
                        res.append(line_num[i] + int2myalphaenum(ls[i][1]) + ' ' + mystr(ls[i]))
                    else:
                        res.append(mystr(ls[i]))

        s = '\n'.join(res)

        # 是否要添加信息头
        if msghead:
            s = head + '\n' + s

        return s
